# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.3 `anova` Chapter.

``anova`` - Analysis of Variance

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module is concerned with methods for analysing the results of
designed experiments.
The range of experiments covered include:

    single factor designs with equal sized blocks such as randomized
    complete block and balanced incomplete block designs,

    row and column designs such as Latin squares, and

    complete factorial designs.

Further designs may be analysed by combining the analyses provided by
multiple calls to functions or by using general linear model functions
provided in submodule ``correg``.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27.3/flhtml/g04/g04intro.html
"""

# NAG Copyright 2017-2021.

import collections as _collections
import logging as _logging
import sys as _sys

from ..base import utils

def hier2(y, n, k, lsub, nobs, l, ngp, gbar, sgbar, ss, idf, f, fp):
    u"""
    Two-way analysis of variance, hierarchical classification, subgroups
    of unequal size.

    ``hier2`` performs an analysis of variance for a two-way
    hierarchical classification with subgroups of possibly unequal size,
    and also computes the treatment group and subgroup means.
    A fixed effects model is assumed.

    For full information please refer to the NAG Library document for
    g04ag

    https://www.nag.com/numeric/nl/nagdoc_27.3/flhtml/g04/g04agf.html

    Parameters
    ----------
    y : float, shape (`n`)
        The elements of `y` must contain the observations y_mij in the
        following order:

            y_111,y_211,...,y_{n_11 11},y_112,y_212,...,y_{n_12
            12},...,y_{11l_1},...,

y_{n_{1l_1}1l_1},...,y_{1ij},...,y_{n_{ij}ij},...,y_{1kl_k},...,y_{n_{kl_k}kl_k}.

        In words, the ordering is by group, and within each group is by
        subgroup, the members of each subgroup being in consecutive
        locations in `y`.

    n : int
        n, the total number of observations.

    k : int
        k, the number of groups.

    lsub : int, shape (`k`)
        The number of subgroups within group i, l_i, for i = 1,2,...,k.

    nobs : int, shape (`l`)
        The numbers of observations in each subgroup, n_{ij}, in the
        following order:

n_11,n_12,...,n_{1l_1},n_21,...,n_{2l_2},...,n_{k1},...,n_{kl_k}

    l : int
        l, the total number of subgroups.

    ngp : int, shape (`k`), modified in place
        `On exit`: the total number of observations in group i, n_{i.},
        for i = 1,2,...,k.

    gbar : float, shape (`k`), modified in place
        `On exit`: the mean for group i, y__{.i.}, for i = 1,2,...,k.

    sgbar : float, shape (`l`), modified in place
        `On exit`: the subgroup means, y__{.ij}, in the following order:

y__.11,y__.12,...,y__.1l_1,y__.21,y__.22,...,y__.2l_2,...,y__{.k1},y__{.k2},...,y__.kl_k.

    ss : float, shape (4), modified in place
        `On exit`: contains the sums of squares for the analysis of
        variance, as follows;

            `ss`[0] = Between group sum of squares, `ss`_g,

            `ss`[1] = Between subgroup within groups sum of squares,
            `ss`_{sg},

            `ss`[2] = Residual sum of squares, `ss`_res,

            `ss`[3] = Corrected total sum of squares, `ss`_tot.

    idf : int, shape (4), modified in place
        `On exit`: contains the degrees of freedom attributable to each
        sum of squares in the analysis of variance, as follows:

            `idf`[0] = Degrees of freedom for between group sum of
            squares,

            `idf`[1] = Degrees of freedom for between subgroup within
            groups sum of squares,

            `idf`[2] = Degrees of freedom for residual sum of squares,

            `idf`[3] = Degrees of freedom for corrected total sum of
            squares.

    f : float, shape (2), modified in place
        `On exit`: contains the mean square ratios, F_1 and F_2, for the
        between groups variation, and the between subgroups within
        groups variation, with respect to the residual, respectively.

    fp : float, shape (2), modified in place
        `On exit`: contains the significances of the mean square ratios,
        p_1 and p_2 respectively.

    Returns
    -------
    gm : float
        The grand mean, y__....

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` > 1.

        (`errno` 2)
            On entry, i = *<value>* and `lsub`[i-1] = *<value>*.

            Constraint: `lsub`[i-1] > 0.

        (`errno` 3)
            On entry, sum_i`lsub`[i] = *<value>* and `l` = *<value>*.

            Constraint: sum_i`lsub`[i] = `l`.

        (`errno` 4)
            On entry, i = *<value>* and `nobs`[i-1] = *<value>*.

            Constraint: `nobs`[i-1] > 0.

        (`errno` 5)
            On entry, sum_i`nobs`[i-1] = *<value>* and `n` = *<value>*.

            Constraint: sum_i`nobs`[i-1] = `n`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            The total corrected sum of squares is zero.

        (`errno` 7)
            The residual sum of squares is zero.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    In a two-way hierarchical classification, there are k (>= 2)
    treatment groups, the ith of which is subdivided into l_i treatment
    subgroups.
    The jth subgroup of group i contains n_{ij} observations, which may
    be denoted by

        y_{1ij},y_{2ij},...,y_{n_{ij}ij}.

    The general observation is denoted by y_{mij}, being the mth
    observation in subgroup j of group i, for 1 <= i <= k, 1 <= j <=
    l_i, 1 <= m <= n_{ij}.

    The following quantities are computed

    (i) The subgroup means

            y__{.ij} = sum_m = 1^n_{ij}y_{mij}/(n_{ij})

    (#) The group means

            y__{.i.} = sum_j = 1^l_isum_m = 1^n_{ij}y_{mij}/sum_j =
            1^l_in_{ij}

    (#) The grand mean

            y__... = sum_i = 1^ksum_j = 1^l_isum_m =
            1^n_{ij}y_{mij}/(sum_i = 1^ksum_j = 1^l_in_{ij})

    (#) The number of observations in each group

            n_{i.} = sum_j = 1^l_in_{ij}

    (#) Sums of squares

        [table omitted]

    (#) Degrees of freedom of variance components

        [table omitted]

        where

            l = sum_i = 1^kl_i,

            n = sum_i = 1^kn_{i.}

    (#) F ratios. These are the ratios of the group and subgroup mean
        squares to the residual mean square.

        [table omitted]

        If either F ratio exceeds 9999.0, the value 9999.0 is assigned
        instead.

    (#) `f` significances. The probability of obtaining a value from the
        appropriate F-distribution which exceeds the computed mean
        square ratio.

        [table omitted]

        where F_{nu_1,nu_2} denotes the central F-distribution with
        degrees of freedom nu_1 and nu_2.

        If any F_i = 9999.0, then p_i is set to zero, i = 1,2.

    References
    ----------
    Kendall, M G and Stuart, A, 1976, `The Advanced Theory of Statistics
    (Volume 3)`, (3rd Edition), Griffin

    Moore, P G, Shirley, E A and Edwards, D E, 1972, `Standard
    Statistical Calculations`, Pitman
    """

    excs = []
    fun_name = 'naginterfaces.base.anova.hier2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g04agft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._EngineIntArrayType.c_type, # Input, lsub[k].
        utils._EngineIntArrayType.c_type, # Input, nobs[l].
        utils._BYREF_INT, # Input, l.
        utils._EngineIntArrayType.c_type, # Output, ngp[k].
        utils._EngineFloat64ArrayType.c_type, # Output, gbar[k].
        utils._EngineFloat64ArrayType.c_type, # Output, sgbar[l].
        utils._BYREF_FLOAT64, # Output, gm.
        utils._EngineFloat64ArrayType.c_type, # Output, ss[4].
        utils._EngineIntArrayType.c_type, # Output, idf[4].
        utils._EngineFloat64ArrayType.c_type, # Output, f[2].
        utils._EngineFloat64ArrayType.c_type, # Output, fp[2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _lsub = utils._EngineIntArrayType(
        lsub, {'fun_name': fun_name, 'entity': 'lsub'},
        exp_shape=(_k_py,),
    ).to_ctypes()
    _l = utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _l_py = utils._EngineIntScalarType.to_py(_l)
    _nobs = utils._EngineIntArrayType(
        nobs, {'fun_name': fun_name, 'entity': 'nobs'},
        exp_shape=(_l_py,),
    ).to_ctypes()
    _ngp_cls = utils._EngineIntArrayType(
        ngp, {'fun_name': fun_name, 'entity': 'ngp'},
        exp_shape=(_k_py,),
        check_contents=False,
    )
    _ngp = _ngp_cls.to_ctypes(can_reference_contents=False)
    _gbar_cls = utils._EngineFloat64ArrayType(
        gbar, {'fun_name': fun_name, 'entity': 'gbar'},
        exp_shape=(_k_py,),
        check_contents=False,
    )
    _gbar = _gbar_cls.to_ctypes(can_reference_contents=False)
    _sgbar_cls = utils._EngineFloat64ArrayType(
        sgbar, {'fun_name': fun_name, 'entity': 'sgbar'},
        exp_shape=(_l_py,),
        check_contents=False,
    )
    _sgbar = _sgbar_cls.to_ctypes(can_reference_contents=False)
    _gm = utils._EngineFloat64ScalarType.empty_ctypes()
    _ss_cls = utils._EngineFloat64ArrayType(
        ss, {'fun_name': fun_name, 'entity': 'ss'},
        exp_shape=(4,),
        check_contents=False,
    )
    _ss = _ss_cls.to_ctypes(can_reference_contents=False)
    _idf_cls = utils._EngineIntArrayType(
        idf, {'fun_name': fun_name, 'entity': 'idf'},
        exp_shape=(4,),
        check_contents=False,
    )
    _idf = _idf_cls.to_ctypes(can_reference_contents=False)
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(2,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _fp_cls = utils._EngineFloat64ArrayType(
        fp, {'fun_name': fun_name, 'entity': 'fp'},
        exp_shape=(2,),
        check_contents=False,
    )
    _fp = _fp_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _y,
        _n,
        _k,
        _lsub,
        _nobs,
        _l,
        _ngp,
        _gbar,
        _sgbar,
        _gm,
        _ss,
        _idf,
        _f,
        _fp,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and lsub[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lsub[i-1] > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sum_ilsub[i] = {:s} and l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: sum_ilsub[i] = l.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and nobs[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nobs[i-1] > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sum_inobs[i-1] = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: sum_inobs[i-1] = n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99994:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The total corrected sum of squares is zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The residual sum of squares is zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ngp_cls.refresh_py(_ngp)
        _gbar_cls.refresh_py(_gbar)
        _sgbar_cls.refresh_py(_sgbar)
        _ss_cls.refresh_py(_ss)
        _idf_cls.refresh_py(_idf)
        _f_cls.refresh_py(_f)
        _fp_cls.refresh_py(_fp)
        _gm_py = utils._EngineFloat64ScalarType.to_py(_gm)
        return_data = _gm_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['gm'],
            )(
                _gm_py,
            )
    _err_class.manage()
    return return_data

def random(n, y, iblock, nt, it, bmean, tmean, tabl, c, irep, r, ef, tol, irdf):
    u"""
    Analysis of variance, randomized block or completely randomized
    design, treatment means and standard errors.

    ``random`` computes the analysis of variance and treatment means and
    standard errors for a randomized block or completely randomized
    design.

    For full information please refer to the NAG Library document for
    g04bb

    https://www.nag.com/numeric/nl/nagdoc_27.3/flhtml/g04/g04bbf.html

    Parameters
    ----------
    n : int
        The number of observations.

    y : float, shape (`n`)
        The observations in the order as described by `iblock` and `nt`.

    iblock : int
        Indicates the block structure.

        abs(`iblock`) <= 1

            There are no blocks, i.e., it is a completely randomized
            design.

        `iblock` >= 2

            There are `iblock` blocks and the data should be input by
            blocks, i.e., `y` must contain the observations for block 1
            followed by the observations for block 2, etc.

        `iblock` <= -2

            There are abs(`iblock`) blocks and the data is input in
            parallel with respect to blocks, i.e., `y`[0] must contain
            the first observation for block 1, `y`[1] must contain the
            first observation for block 2...`y`[abs(`iblock`)-1] must
            contain the first observation for block
            abs(`iblock`),`y`[abs(`iblock`+1)-1] must contain the second
            observation for block 1, etc.

    nt : int
        The number of treatments. If only blocks are required in the
        analysis then set `nt` = 1.

    it : int, shape (:)
        Note: the required length for this argument is determined as
        follows: if `nt`>=2: `n`; otherwise: 1.

        `it`[i-1] indicates which of the `nt` treatments plot i
        received, for i = 1,2,...,`n`.

        If `nt` = 1, `it` is not referenced.

    bmean : float, shape (|`iblock`|), modified in place
        `On exit`: if abs(`iblock`) >= 2, `bmean`[j-1] contains the mean
        for the jth block, beta^_j, for j = 1,2,...,b.

    tmean : float, shape (`nt`), modified in place
        `On exit`: if `nt` >= 2, `tmean`[l-1] contains the (adjusted)
        mean for the lth treatment, mu^*+tau^_l, for l = 1,2,...,t,
        where mu^* is the mean of the treatment adjusted observations,
        y_{ij(l)}-tau^_l.

    tabl : float, shape (4, 5), modified in place
        `On exit`: the analysis of variance table. Column 1 contains the
        degrees of freedom, column 2 the sum of squares, and where
        appropriate, column 3 the mean squares, column 4 the F-statistic
        and column 5 the significance level of the F-statistic. Row 1 is
        for Blocks, row 2 for Treatments, row 3 for Residual and row 4
        for Total. Mean squares are computed for all but the Total row;
        F-statistics and significance are computed for Treatments and
        Blocks, if present. Any unfilled cells are set to zero.

    c : float, shape (`nt`, `nt`), modified in place
        `On exit`: if `nt` >= 2, the upper triangular part of `c`
        contains the variance-covariance matrix of the treatment
        effects, the strictly lower triangular part contains the
        standard errors of the difference between two treatment effects
        (means), i.e., `c`[i-1,j-1] contains the covariance of treatment
        i and j if j >= i and the standard error of the difference
        between treatment i and j if j < i, for j = 1,2,...,t, for i =
        1,2,...,t.

    irep : int, shape (`nt`), modified in place
        `On exit`: if `nt` >= 2, the treatment replications, R_{ll}, for
        l = 1,2,...,`nt`.

    r : float, shape (`n`), modified in place
        `On exit`: the residuals, r_i, for i = 1,2,...,`n`.

    ef : float, shape (`nt`), modified in place
        `On exit`: if `nt` >= 2, the canonical efficiency factors.

    tol : float
        The tolerance value used to check for zero eigenvalues of the
        matrix Omega. If `tol` = 0.0 a default value of 10^-5 is used.

    irdf : int
        An adjustment to the degrees of freedom for the residual and
        total.

        `irdf` >= 1

            The degrees of freedom for the total is set to `n`-`irdf`
            and the residual degrees of freedom adjusted accordingly.

        `irdf` = 0

            The total degrees of freedom for the total is set to `n`-1,
            as usual.

    Returns
    -------
    gmean : float
        The grand mean, mu^.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `irdf` = *<value>*.

            Constraint: `irdf` >= 0.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, no blocks or treatments in model.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `n` = *<value>* and `iblock` = *<value>*.

            Constraint: `n` >= 2 and if abs(`iblock`) >= 2, `n` must be
            a multiple of abs(`iblock`).

        (`errno` 3)
            On entry, at least one treatment is not present. Treatment
            *<value>* does not appear in `it`.

        (`errno` 3)
            On entry, i = *<value>*, `it`[i-1] = *<value>* and `nt` =
            *<value>*.

            Constraint: 1 <= `it`[i-1] <= `nt`.

        (`errno` 4)
            On entry, the values in `y` are constant.

        (`errno` 5)
            The computation of the eigenvalues has failed to converge.

        (`errno` 5)
            A computed standard error is zero.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            The treatments are totally confounded with blocks.

        (`errno` 7)
            The residual degrees of freedom is zero.

        (`errno` 7)
            The residual mean square is zero.

        (`errno` 8)
            The design is disconnected.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    In a completely randomized design, experimental material is divided
    into a number of units, or plots, to which a treatment can be
    applied.
    In a randomized block design the units are grouped into blocks so
    that the variation within blocks is less than the variation between
    blocks.
    If every treatment is applied to one plot in each block it is a
    complete block design.
    If there are fewer plots per block than treatments then the design
    will be an incomplete block design and may be balanced or partially
    balanced.

    For a completely randomized design, with t treatments and n_t plots
    per treatment, the linear model is

        y_{ij} = mu+tau_j+e_{ij}, j = 1,2,...,t and i = 1,2,...,n_j,

    where y_{ij} is the ith observation for the jth treatment, mu is the
    overall mean, tau_j is the effect of the jth treatment and e_{ij} is
    the random error term.
    For a randomized block design, with t treatments and b blocks of k
    plots, the linear model is

        y_{ij(l)} = mu+beta_i+tau_l+e_{ij}, i = 1,2,...,b, j = 1,2,...,k
        and l = 1,2,...,t,

    where beta_i is the effect of the ith block and the ij(l) notation
    indicates that the lth treatment is applied to the jth plot in the
    ith block.

    The completely randomized design gives rise to a one-way analysis of
    variance.
    The treatments do not have to be equally replicated, i.e., do not
    have to occur the same number of times.
    First the overall mean, mu^, is computed and subtracted from the
    observations to give y'_ij = y_{ij}-mu^.
    The estimated treatment effects, tau^_j are then computed as the
    treatment means of the mean adjusted observations, y'_ij, and the
    treatment sum of squares can be computed from the sum of squares of
    the treatment totals of the y'_ij divided by the number of
    observations per treatment total, n_j.
    The final residuals are computed as r_{ij} = y'_ij-tau^_j, and, from
    the residuals, the residual sum of squares is calculated.

    For the randomized block design the mean is computed and subtracted
    from the observations to give y'_ij(l) = y_{ij(l)}-mu^.
    The estimated block effects, ignoring treatment effects, beta^_i,
    are then computed using the block means of the y'_ij(l) and the
    unadjusted sum of squares computed as the sum of squared block
    totals for the y'_ij(l) divided by number of plots per block, k.
    The block adjusted observations are then computed as y_ij(l)^'' =
    y'_ij_(l) = beta^_i.
    In the case of the complete block design, with the same replication
    for each treatment within each block, the blocks and treatments are
    orthogonal, and so the treatment effects are estimated as the
    treatment means of the block adjusted observations, y_ij(l)^''.
    The treatment sum of squares is computed as the sum of squared
    treatment totals of the y_ij(l)^'' divided by the number of
    replicates to the treatments, r = bk/t.
    Finally the residuals, and hence the residual sum of squares, are
    given by r_{ij(l)} = y_ij(l)^''-tau^_l.

    For a design without the same replication for each treatment within
    each block the treatments and the blocks will not be orthogonal, so
    the treatments adjusted for blocks need to be computed.
    The adjusted treatment effects are found as the solution to the
    equations

        (R-NN^T/k)tau^= q,

    where q is the vector of the treatment totals for block adjusted
    observations, y_ij(l)^'', R is a diagonal matrix with R_{ll} equal
    to the number of times the lth treatment is replicated, and `n` is
    the t*b incidence matrix, with N_{lj} equal to the number of times
    treatment l occurs in block j.
    The solution to the equations can be written as

        tau^= Omega q

    where Omega is a generalized inverse of (R-NN^T/k).
    The solution is found from the eigenvalue decomposition of
    (R-NN^T/k).
    The residuals are first calculated by subtracting the estimated
    treatment effects from the block adjusted observations to give
    r'_ij(l) = y_ij(l)^''-tau^_l.
    However, since only the unadjusted block effects have been removed
    and blocks and treatments are not orthogonal, the block means of the
    r'_ij(l) have to be subtracted to give the correct residuals,
    r_{ij(l)} and residual sum of squares.

    The mean squares are computed as the sum of squares divided by the
    degrees of freedom.
    The degrees of freedom for the unadjusted blocks is b-1, for the
    completely randomized and the complete block designs the degrees of
    freedom for the treatments is t-1.
    In the general case the degrees of freedom for treatments is the
    rank of the matrix Omega.
    The F-statistic given by the ratio of the treatment mean square to
    the residual mean square tests the hypothesis

        H_0:tau_1 = tau_2 = ... = tau_t = 0.

    The standard errors for the difference in treatment effects, or
    treatment means, for the completely randomized or the complete block
    designs, are given by:

        se(tau_j-tau_{j*}) = (1/(n_j)+1/(n_{j*}))s^2

    where s^2 is the residual mean square and n_j = n_{j*} = b in the
    complete block design.
    In the general case the variances of the treatment effects are given
    by

        var(tau) = Omega s^2

    from which the appropriate standard errors of the difference between
    treatment effects or the difference between adjusted means can be
    calculated.

    In the complete block design all the information on the treatment
    effects is given by the within block analysis.
    In other designs there may be a loss of information due to the
    non-orthogonality of treatments and blocks.
    The efficiency of the within block analysis in these cases is given
    by the (canonical) efficiency factors, these are the nonzero
    eigenvalues of the matrix (R-NN^T/k), divided by the number of
    replicates in the case of equal replication, or by the mean of the
    number of replicates in the unequally replicated case, see John
    (1987).
    If more than one eigenvalue is zero then the design is said to be
    disconnected and some treatments can only be compared using a
    between block analysis.

    References
    ----------
    Cochran, W G and Cox, G M, 1957, `Experimental Designs`, Wiley

    Davis, O L, 1978, `The Design and Analysis of Industrial
    Experiments`, Longman

    John, J A, 1987, `Cyclic Designs`, Chapman and Hall

    John, J A and Quenouille, M H, 1977, `Experiments: Design and
    Analysis`, Griffin

    Searle, S R, 1971, `Linear Models`, Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.anova.random'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g04bbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._BYREF_INT, # Input, iblock.
        utils._BYREF_INT, # Input, nt.
        utils._EngineIntArrayType.c_type, # Input, it[*].
        utils._BYREF_FLOAT64, # Output, gmean.
        utils._EngineFloat64ArrayType.c_type, # Output, bmean[abs(iblock)].
        utils._EngineFloat64ArrayType.c_type, # Output, tmean[nt].
        utils._EngineFloat64ArrayType.c_type, # Output, tabl[ldtabl, 5].
        utils._BYREF_INT, # Input, ldtabl.
        utils._EngineFloat64ArrayType.c_type, # Output, c[ldc, nt].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineIntArrayType.c_type, # Output, irep[nt].
        utils._EngineFloat64ArrayType.c_type, # Output, r[n].
        utils._EngineFloat64ArrayType.c_type, # Output, ef[nt].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, irdf.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((tabl, 'tabl'), (c, 'c')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _iblock = utils._EngineIntScalarType(
        iblock, {'fun_name': fun_name, 'entity': 'iblock'},
    ).to_ctypes()
    _nt = utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _nt_py = utils._EngineIntScalarType.to_py(_nt)
    if _nt_py >= 2:
        _it_dim1 = _n_py
    else:
        _it_dim1 = 1
    exp_shape_it = (_it_dim1,)
    _it = utils._EngineIntArrayType(
        it, {'fun_name': fun_name, 'entity': 'it'},
        exp_shape=exp_shape_it,
    ).to_ctypes()
    _gmean = utils._EngineFloat64ScalarType.empty_ctypes()
    _iblock_py = utils._EngineIntScalarType.to_py(_iblock)
    _bmean_cls = utils._EngineFloat64ArrayType(
        bmean, {'fun_name': fun_name, 'entity': 'bmean'},
        exp_shape=(abs(_iblock_py),),
        check_contents=False,
    )
    _bmean = _bmean_cls.to_ctypes(can_reference_contents=False)
    _tmean_cls = utils._EngineFloat64ArrayType(
        tmean, {'fun_name': fun_name, 'entity': 'tmean'},
        exp_shape=(_nt_py,),
        check_contents=False,
    )
    _tmean = _tmean_cls.to_ctypes(can_reference_contents=False)
    _ldtabl_py = utils._EngineFloat64ArrayType.get_nd(tabl, {'fun_name': fun_name, 'entity': 'tabl'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldtabl = utils._EngineIntScalarType(
        _ldtabl_py, {'fun_name': fun_name, 'entity': '_ldtabl'},
    ).to_ctypes()
    _tabl_cls = utils._EngineFloat64ArrayType(
        tabl, {'fun_name': fun_name, 'entity': 'tabl'},
        exp_shape=(4, 5),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _tabl = _tabl_cls.to_ctypes(can_reference_contents=False)
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_nt_py, _nt_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _irep_cls = utils._EngineIntArrayType(
        irep, {'fun_name': fun_name, 'entity': 'irep'},
        exp_shape=(_nt_py,),
        check_contents=False,
    )
    _irep = _irep_cls.to_ctypes(can_reference_contents=False)
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _r = _r_cls.to_ctypes(can_reference_contents=False)
    _ef_cls = utils._EngineFloat64ArrayType(
        ef, {'fun_name': fun_name, 'entity': 'ef'},
        exp_shape=(_nt_py,),
        check_contents=False,
    )
    _ef = _ef_cls.to_ctypes(can_reference_contents=False)
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _irdf = utils._EngineIntScalarType(
        irdf, {'fun_name': fun_name, 'entity': 'irdf'},
    ).to_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _y,
        _iblock,
        _nt,
        _it,
        _gmean,
        _bmean,
        _tmean,
        _tabl,
        _ldtabl,
        _c,
        _ldc,
        _irep,
        _r,
        _ef,
        _tol,
        _irdf,
        _wk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irdf = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irdf >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, no blocks or treatments in model.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nt >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and iblock = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n >= 2 and if abs(iblock) >= 2,',
                    'n must be a multiple of abs(iblock).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one treatment is not present.',
                    'Treatment {:s} does not appear in it.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, it[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= it[i-1] <= nt.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the values in y are constant.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The computation of the eigenvalues has failed to converge.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A computed standard error is zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99983:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The treatments are totally confounded with blocks.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99984:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The residual degrees of freedom is zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99985:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The residual mean square is zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99987:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The design is disconnected.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _bmean_cls.refresh_py(_bmean)
        _tmean_cls.refresh_py(_tmean)
        _tabl_cls.refresh_py(_tabl)
        _c_cls.refresh_py(_c)
        _irep_cls.refresh_py(_irep)
        _r_cls.refresh_py(_r)
        _ef_cls.refresh_py(_ef)
        _gmean_py = utils._EngineFloat64ScalarType.to_py(_gmean)
        return_data = _gmean_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['gmean'],
            )(
                _gmean_py,
            )
    _err_class.manage()
    return return_data

def rowcol(nrep, nrow, ncol, y, nt, it, tmean, tabl, c, irep, rpmean, rmean, cmean, r, ef, tol, irdf):
    u"""
    Analysis of variance, general row and column design, treatment means
    and standard errors.

    ``rowcol`` computes the analysis of variance for a general row and
    column design together with the treatment means and standard errors.

    For full information please refer to the NAG Library document for
    g04bc

    https://www.nag.com/numeric/nl/nagdoc_27.3/flhtml/g04/g04bcf.html

    Parameters
    ----------
    nrep : int
        b, the number of replicates.

    nrow : int
        r, the number of rows per replicate.

    ncol : int
        c, the number of columns per replicate.

    y : float, shape (`nrep`*`nrow`*`ncol`)
        The n = brc observations ordered by columns within rows within
        replicates. That is `y`[ rc(i-1) +r(j-1) +k -1] contains the
        observation from the kth column of the jth row of the ith
        replicate, for k = 1,2,...,c, for j = 1,2,...,r, for i =
        1,2,...,b.

    nt : int
        The number of treatments. If only replicates, rows and columns
        are required in the analysis then set `nt` = 1.

    it : int, shape (:)
        Note: the required length for this argument is determined as
        follows: if `nt`>1: `nrep`*`nrow`*`ncol`; otherwise: 1.

        If `nt` > 1, `it`[i-1] indicates which of the `nt` treatments
        unit i received, for i = 1,2,...,n.

        If `nt` = 1, `it` is not referenced.

    tmean : float, shape (`nt`), modified in place
        `On exit`: if `nt` >= 2, `tmean`[l-1] contains the (adjusted)
        mean for the lth treatment, mu^*+tau^_l, for l = 1,2,...,t,
        where mu^* is the mean of the treatment adjusted observations
        y_{ijk(l)}-tau^_l. Otherwise `tmean` is not referenced.

    tabl : float, shape (6, 5), modified in place
        `On exit`: the analysis of variance table. Column 1 contains the
        degrees of freedom, column 2 the sum of squares, and where
        appropriate, column 3 the mean squares, column 4 the F-statistic
        and column 5 the significance level of the F-statistic. Row 1 is
        for replicates, row 2 for rows, row 3 for columns, row 4 for
        treatments (if `nt` > 1), row 5 for residual and row 6 for
        total. Mean squares are computed for all but the total row,
        F-statistics and significance are computed for treatments,
        replicates, rows and columns. Any unfilled cells are set to
        zero.

    c : float, shape (`nt`, `nt`), modified in place
        `On exit`: the upper triangular part of `c` contains the
        variance-covariance matrix of the treatment effects, the
        strictly lower triangular part contains the standard errors of
        the difference between two treatment effects (means), i.e.,
        `c`[i-1,j-1] contains the covariance of treatment i and j if j
        >= i and the standard error of the difference between treatment
        i and j if j < i, for j = 1,2,...,t, for i = 1,2,...,t.

    irep : int, shape (`nt`), modified in place
        `On exit`: if `nt` > 1, the treatment replications, R_{ll}, for
        l = 1,2,...,`nt`. Otherwise `irep` is not referenced.

    rpmean : float, shape (`nrep`), modified in place
        `On exit`: if `nrep` > 1, `rpmean`[i-1] contains the mean for
        the ith replicate, mu^+beta^_i, for i = 1,2,...,b. Otherwise
        `rpmean` is not referenced.

    rmean : float, shape (`nrep`*`nrow`), modified in place
        `On exit`: `rmean`[j-1] contains the mean for the jth row,
        mu^+rho^_i, for j = 1,2,...,r.

    cmean : float, shape (`nrep`*`ncol`), modified in place
        `On exit`: `cmean`[k-1] contains the mean for the kth column,
        mu^+gamma^_k, for k = 1,2,...,c.

    r : float, shape (`nrep`*`nrow`*`ncol`), modified in place
        `On exit`: the residuals, r_i, for i = 1,2,...,n.

    ef : float, shape (`nt`), modified in place
        `On exit`: if `nt` >= 2, the canonical efficiency factors.
        Otherwise `ef` is not referenced.

    tol : float
        The tolerance value used to check for zero eigenvalues of the
        matrix Omega. If `tol` = 0.0 a default value of 0.00001 is used.

    irdf : int
        An adjustment to the degrees of freedom for the residual and
        total.

        `irdf` >= 1

            The degrees of freedom for the total is set to n-`irdf` and
            the residual degrees of freedom adjusted accordingly.

        `irdf` = 0

            The total degrees of freedom for the total is set to n-1, as
            usual.

    Returns
    -------
    gmean : float
        The grand mean, mu^.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nrep` = *<value>*.

            Constraint: `nrep` >= 1.

        (`errno` 1)
            On entry, `irdf` = *<value>*.

            Constraint: `irdf` >= 0.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, `ncol` = *<value>*.

            Constraint: `ncol` >= 2.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` >= 1.

        (`errno` 1)
            On entry, `nrow` = *<value>*.

            Constraint: `nrow` >= 2.

        (`errno` 2)
            On entry, at least one treatment is not present. Treatment
            *<value>* does not appear in `it`.

        (`errno` 2)
            On entry, i = *<value>*, `it`[i-1] = *<value>* and `nt` =
            *<value>*.

            Constraint: 1 <= `it`[i-1] <= `nt`.

        (`errno` 3)
            On entry, the values in `y` are constant.

        (`errno` 4)
            The computation of the eigenvalues has failed to converge.

        (`errno` 4)
            A computed standard error is zero.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The treatments are totally confounded with blocks.

        (`errno` 6)
            The residual degrees of freedom is zero.

        (`errno` 6)
            The residual mean square is zero.

        (`errno` 7)
            The design is disconnected.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    In a row and column design the experimental material can be
    characterised by a two-way classification, nominally called rows and
    columns.
    Each experimental unit can be considered as being located in a
    particular row and column.
    It is assumed that all rows are of the same length and all columns
    are of the same length.
    Sets of equal numbers of rows/columns can be grouped together to
    form replicates, sometimes known as squares or rectangles, as
    appropriate.

    If for a replicate, the number of rows, the number of columns and
    the number of treatments are equal and every treatment occurs once
    in each row and each column then the design is a Latin square.
    If this is not the case the treatments will be non-orthogonal to
    rows and columns.
    For example in the case of a lattice square each treatment occurs
    only once in each square.

    For a row and column design, with t treatments in r rows and c
    columns and b replicates or squares with n = brc observations the
    linear model is:

        y_{ijk(l)} = mu+beta_i+rho_j+gamma_k+tau_l+e_{ijk}

    for i = 1,2...,b, j = 1,2,...,r, k = 1,2...,c and l = 1,2,...,t,
    where beta_i is the effect of the ith replicate, rho_j is the effect
    of the jth row, gamma_k is the effect of the kth column and the
    ijk(l) notation indicates that the lth treatment is applied to the
    unit in row j, column k of replicate i.

    To compute the analysis of variance for a row and column design the
    mean is computed and subtracted from the observations to give,
    y'_ijk(l) = y_{ijk(l)}-mu^.
    Since the replicates, rows and columns are orthogonal the estimated
    effects, ignoring treatment effects, beta^_i, rho^_j, gamma^_k, can
    be computed using the appropriate means of the y'_ijk(l), and the
    unadjusted sum of squares computed as the appropriate sum of squared
    totals for the y'_ijk(l) divided by number of units per total.
    The observations adjusted for replicates, rows and columns can then
    be computed by subtracting the estimated effects from y'_ijk(l) to
    give y_ijk(l)^''.

    In the case of a Latin square design the treatments are orthogonal
    to replicates, rows and columns and so the treatment effects,
    tau^_l, can be estimated as the treatment means of the adjusted
    observations, y_ijk(l)^''.
    The treatment sum of squares is computed as the sum of squared
    treatment totals of the y_ij(l)^'' divided by the number of times
    each treatment is replicated.
    Finally the residuals, and hence the residual sum of squares, are
    given by r_{ij(l)} = y_ij(l)^''-tau^_l.

    For a design which is not orthogonal, for example a lattice square
    or an incomplete Latin square, the treatment effects adjusted for
    replicates, rows and columns need to be computed.
    The adjusted treatment effects are found as the solution to the
    equations:

        A tau^= (R-N_b N_b^T/(rc)-N_r N_r^T/(bc)-N_c N_c^T/(br))tau^= q

    where q is the vector of the treatment totals of the observations
    adjusted for replicates, rows and columns, y_ijk(l)^'', R is a
    diagonal matrix with R_{ll} equal to the number of times the lth
    treatment is replicated, and N_b is the t*b incidence matrix, with
    N_{l,i} equal to the number of times treatment l occurs in replicate
    i, with N_r and N_c being similarly defined for rows and columns.
    The solution to the equations can be written as:

        tau^= Omega q

    where, Omega is a generalized inverse of A.
    The solution is found from the eigenvalue decomposition of A.
    The residuals are first calculated by subtracting the estimated
    adjusted treatment effects from the adjusted observations to give
    r'_ij(l) = y_ij(l)^''-tau^_l.
    However, since only the unadjusted replicate, row and column effects
    have been removed and they are not orthogonal to treatments, the
    replicate, row and column means of the r'_ij(l) have to be
    subtracted to give the correct residuals, r_{ij(l)} and residual sum
    of squares.

    Given the sums of squares, the mean squares are computed as the sums
    of squares divided by the degrees of freedom.
    The degrees of freedom for the unadjusted replicates, rows and
    columns are b-1, r-1 and c-1 respectively and for the Latin square
    designs the degrees of freedom for the treatments is t-1.
    In the general case the degrees of freedom for treatments is the
    rank of the matrix Omega.
    The F-statistic given by the ratio of the treatment mean square to
    the residual mean square tests the hypothesis:

        H_0:tau_1 = tau_2 = ... = tau_t = 0.

    The standard errors for the difference in treatment effects, or
    treatment means, for Latin square designs, are given by:

        se(tau^_j-tau^_{j*}) = sqrt(2s^2/(bt))

    where s^2 is the residual mean square.
    In the general case the variances of the treatment effects are given
    by:

        Var(tau^) = Omega s^2

    from which the appropriate standard errors of the difference between
    treatment effects or the difference between adjusted means can be
    calculated.

    The analysis of a row-column design can be considered as consisting
    of different strata: the replicate stratum, the rows within
    replicate and the columns within replicate strata and the units
    stratum.
    In the Latin square design all the information on the treatment
    effects is given at the units stratum.
    In other designs there may be a loss of information due to the
    non-orthogonality of treatments and replicates, rows and columns and
    information on treatments may be available in higher strata.
    The efficiency of the estimation at the units stratum is given by
    the (canonical) efficiency factors, these are the nonzero
    eigenvalues of the matrix, A, divided by the number of replicates in
    the case of equal replication, or by the mean of the number of
    replicates in the unequally replicated case, see John (1987).
    If more than one eigenvalue is zero then the design is said to be
    disconnected and information on some treatment comparisons can only
    be obtained from higher strata.

    References
    ----------
    Cochran, W G and Cox, G M, 1957, `Experimental Designs`, Wiley

    Davis, O L, 1978, `The Design and Analysis of Industrial
    Experiments`, Longman

    John, J A, 1987, `Cyclic Designs`, Chapman and Hall

    John, J A and Quenouille, M H, 1977, `Experiments: Design and
    Analysis`, Griffin

    Searle, S R, 1971, `Linear Models`, Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.anova.rowcol'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g04bcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nrep.
        utils._BYREF_INT, # Input, nrow.
        utils._BYREF_INT, # Input, ncol.
        utils._EngineFloat64ArrayType.c_type, # Input, y[nrep*nrow*ncol].
        utils._BYREF_INT, # Input, nt.
        utils._EngineIntArrayType.c_type, # Input, it[*].
        utils._BYREF_FLOAT64, # Output, gmean.
        utils._EngineFloat64ArrayType.c_type, # Output, tmean[nt].
        utils._EngineFloat64ArrayType.c_type, # Output, tabl[ldtabl, 5].
        utils._BYREF_INT, # Input, ldtabl.
        utils._EngineFloat64ArrayType.c_type, # Output, c[ldc, nt].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineIntArrayType.c_type, # Output, irep[nt].
        utils._EngineFloat64ArrayType.c_type, # Output, rpmean[nrep].
        utils._EngineFloat64ArrayType.c_type, # Output, rmean[nrep*nrow].
        utils._EngineFloat64ArrayType.c_type, # Output, cmean[nrep*ncol].
        utils._EngineFloat64ArrayType.c_type, # Output, r[nrep*nrow*ncol].
        utils._EngineFloat64ArrayType.c_type, # Output, ef[nt].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, irdf.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((tabl, 'tabl'), (c, 'c')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _nrep = utils._EngineIntScalarType(
        nrep, {'fun_name': fun_name, 'entity': 'nrep'},
    ).to_ctypes()
    _nrow = utils._EngineIntScalarType(
        nrow, {'fun_name': fun_name, 'entity': 'nrow'},
    ).to_ctypes()
    _ncol = utils._EngineIntScalarType(
        ncol, {'fun_name': fun_name, 'entity': 'ncol'},
    ).to_ctypes()
    _nrep_py = utils._EngineIntScalarType.to_py(_nrep)
    _nrow_py = utils._EngineIntScalarType.to_py(_nrow)
    _ncol_py = utils._EngineIntScalarType.to_py(_ncol)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_nrep_py*_nrow_py*_ncol_py,),
    ).to_ctypes()
    _nt = utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _nt_py = utils._EngineIntScalarType.to_py(_nt)
    if _nt_py > 1:
        _it_dim1 = _nrep_py*_nrow_py*_ncol_py
    else:
        _it_dim1 = 1
    exp_shape_it = (_it_dim1,)
    _it = utils._EngineIntArrayType(
        it, {'fun_name': fun_name, 'entity': 'it'},
        exp_shape=exp_shape_it,
    ).to_ctypes()
    _gmean = utils._EngineFloat64ScalarType.empty_ctypes()
    _tmean_cls = utils._EngineFloat64ArrayType(
        tmean, {'fun_name': fun_name, 'entity': 'tmean'},
        exp_shape=(_nt_py,),
        check_contents=False,
    )
    _tmean = _tmean_cls.to_ctypes(can_reference_contents=False)
    _ldtabl_py = utils._EngineFloat64ArrayType.get_nd(tabl, {'fun_name': fun_name, 'entity': 'tabl'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldtabl = utils._EngineIntScalarType(
        _ldtabl_py, {'fun_name': fun_name, 'entity': '_ldtabl'},
    ).to_ctypes()
    _tabl_cls = utils._EngineFloat64ArrayType(
        tabl, {'fun_name': fun_name, 'entity': 'tabl'},
        exp_shape=(6, 5),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _tabl = _tabl_cls.to_ctypes(can_reference_contents=False)
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_nt_py, _nt_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _irep_cls = utils._EngineIntArrayType(
        irep, {'fun_name': fun_name, 'entity': 'irep'},
        exp_shape=(_nt_py,),
        check_contents=False,
    )
    _irep = _irep_cls.to_ctypes(can_reference_contents=False)
    _rpmean_cls = utils._EngineFloat64ArrayType(
        rpmean, {'fun_name': fun_name, 'entity': 'rpmean'},
        exp_shape=(_nrep_py,),
        check_contents=False,
    )
    _rpmean = _rpmean_cls.to_ctypes(can_reference_contents=False)
    _rmean_cls = utils._EngineFloat64ArrayType(
        rmean, {'fun_name': fun_name, 'entity': 'rmean'},
        exp_shape=(_nrep_py*_nrow_py,),
        check_contents=False,
    )
    _rmean = _rmean_cls.to_ctypes(can_reference_contents=False)
    _cmean_cls = utils._EngineFloat64ArrayType(
        cmean, {'fun_name': fun_name, 'entity': 'cmean'},
        exp_shape=(_nrep_py*_ncol_py,),
        check_contents=False,
    )
    _cmean = _cmean_cls.to_ctypes(can_reference_contents=False)
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_nrep_py*_nrow_py*_ncol_py,),
        check_contents=False,
    )
    _r = _r_cls.to_ctypes(can_reference_contents=False)
    _ef_cls = utils._EngineFloat64ArrayType(
        ef, {'fun_name': fun_name, 'entity': 'ef'},
        exp_shape=(_nt_py,),
        check_contents=False,
    )
    _ef = _ef_cls.to_ctypes(can_reference_contents=False)
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _irdf = utils._EngineIntScalarType(
        irdf, {'fun_name': fun_name, 'entity': 'irdf'},
    ).to_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nrep,
        _nrow,
        _ncol,
        _y,
        _nt,
        _it,
        _gmean,
        _tmean,
        _tabl,
        _ldtabl,
        _c,
        _ldc,
        _irep,
        _rpmean,
        _rmean,
        _cmean,
        _r,
        _ef,
        _tol,
        _irdf,
        _wk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrep = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrep >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irdf = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irdf >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ncol >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nt >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrow = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrow >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one treatment is not present.',
                    'Treatment {:s} does not appear in it.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, it[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= it[i-1] <= nt.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the values in y are constant.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The computation of the eigenvalues has failed to converge.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A computed standard error is zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99983:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The treatments are totally confounded with blocks.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99984:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The residual degrees of freedom is zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99985:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The residual mean square is zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99987:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The design is disconnected.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _tmean_cls.refresh_py(_tmean)
        _tabl_cls.refresh_py(_tabl)
        _c_cls.refresh_py(_c)
        _irep_cls.refresh_py(_irep)
        _rpmean_cls.refresh_py(_rpmean)
        _rmean_cls.refresh_py(_rmean)
        _cmean_cls.refresh_py(_cmean)
        _r_cls.refresh_py(_r)
        _ef_cls.refresh_py(_ef)
        _gmean_py = utils._EngineFloat64ScalarType.to_py(_gmean)
        return_data = _gmean_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['gmean'],
            )(
                _gmean_py,
            )
    _err_class.manage()
    return return_data

def factorial(n, y, nfac, lfac, nblock, inter, irdf, mterm, table, tmean, maxt, e, imean, semean, bmean, r):
    u"""
    Analysis of variance, complete factorial design, treatment means and
    standard errors.

    ``factorial`` computes an analysis of variance table and treatment
    means for a complete factorial design.

    For full information please refer to the NAG Library document for
    g04ca

    https://www.nag.com/numeric/nl/nagdoc_27.3/flhtml/g04/g04caf.html

    Parameters
    ----------
    n : int
        The number of observations.

    y : float, shape (`n`)
        The observations in standard order, see [Notes].

    nfac : int
        k, the number of factors.

    lfac : int, shape (`nfac`)
        `lfac`[i] must contain the number of levels for the ith factor,
        for i = 1,2,...,k.

    nblock : int
        The number of blocks. If there are no blocks, set `nblock` = 0
        or 1.

    inter : int
        The maximum number of factors in an interaction term. If no
        interaction terms are to be computed, set `inter` = 0 or 1.

    irdf : int
        The adjustment to the residual and total degrees of freedom. The
        total degrees of freedom are set to `n`-`irdf` and the residual
        degrees of freedom adjusted accordingly. For examples of the use
        of `irdf` see [Further Comments].

    mterm : int
        The maximum number of terms in the analysis of variance table,
        see [Further Comments].

    table : float, shape (`mterm`, 5), modified in place
        `On exit`: the first `itotal` rows of `table` contain the
        analysis of variance table. The first column contains the
        degrees of freedom, the second column contains the sum of
        squares, the third column (except for the row corresponding to
        the total sum of squares) contains the mean squares, i.e., the
        sums of squares divided by the degrees of freedom, and the
        fourth and fifth columns contain the F ratio and significance
        level, respectively (except for rows corresponding to the total
        sum of squares, and the residual sum of squares). All other
        cells of the table are set to zero.

        The first row corresponds to the blocks and is set to zero if
        there are no blocks.

        The `itotal`th row corresponds to the total sum of squares for
        `y` and the (`itotal`-1)th row corresponds to the residual sum
        of squares.

        The central rows of the table correspond to the main effects
        followed by the interaction if specified by `inter`.

        The main effects are in the order specified by `lfac` and the
        interactions are in lexical order, see [Notes].

    tmean : float, shape (`maxt`), modified in place
        `On exit`: the treatment means. The position of the means for an
        effect is given by the index in `imean`. For a given effect the
        means are in standard order, see [Notes].

    maxt : int
        The maximum number of treatment means to be computed, see
        [Further Comments]. If the value of `maxt` is too small for the
        required analysis then the minimum number is returned in
        `imean`[0].

    e : float, shape (`maxt`), modified in place
        `On exit`: the estimated effects in the same order as for the
        means in `tmean`.

    imean : int, shape (`mterm`), modified in place
        `On exit`: indicates the position of the effect means in
        `tmean`. The effect means corresponding to the first treatment
        effect in the ANOVA table are stored in `tmean`[0] up to
        `tmean`[`imean`[0]-1]. Other effect means corresponding to the
        ith treatment effect, i = 1,2,...,`itotal`-3, are stored in
        `tmean`[`imean`[i-1]] up to `tmean`[`imean`[i]-1].

    semean : float, shape (`mterm`), modified in place
        `On exit`: the standard error of the difference between means
        corresponding to the ith treatment effect in the ANOVA table.

    bmean : float, shape (`nblock`+1), modified in place
        `On exit`: `bmean`[0] contains the grand mean, if `nblock` > 1,
        `bmean`[1] up to `bmean`[`nblock`] contain the block means.

    r : float, shape (`n`), modified in place
        `On exit`: the residuals.

    Returns
    -------
    itotal : int
        The row in `table` corresponding to the total sum of squares.
        The number of treatment effects is `itotal`-3.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `irdf` = *<value>*.

            Constraint: `irdf` >= 0.

        (`errno` 1)
            On entry, `inter` = *<value>* and `nfac` = *<value>*.

            Constraint: `inter` <= `nfac`.

        (`errno` 1)
            On entry, `inter` = *<value>*.

            Constraint: `inter` >= 0.

        (`errno` 1)
            On entry, `nblock` = *<value>*.

            Constraint: `nblock` >= 0.

        (`errno` 1)
            On entry, `nfac` = *<value>*.

            Constraint: `nfac` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 4.

        (`errno` 2)
            On entry, i = *<value>* and `lfac`[i-1] = *<value>*

            Constraint: `lfac`[i-1] >= 2.

        (`errno` 2)
            On entry, the number of plots per block is incompatible with
            the number of plot factors.

        (`errno` 2)
            On entry, `n` = *<value>* and `nblock` = *<value>*.

            Constraint: when `nblock` > 1, `n` must be a multiple of
            `nblock`.

        (`errno` 3)
            On entry, the values of `y` are constant.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The residual sum of squares is zero.

        (`errno` 4)
            There are no degrees of freedom for the residual sum of
            squares.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    An experiment consists of a collection of units, or plots, to which
    a number of treatments are applied.
    In a factorial experiment the effects of several different sets of
    conditions are compared, e.g., three different temperatures, T_1,
    T_2 and T_3, and two different pressures, P_1 and P_2.
    The conditions are known as factors and the different values the
    conditions take are known as levels.
    In a factorial experiment the experimental treatments are the
    combinations of all the different levels of all factors, e.g.,

        [table omitted]

    The effect of a factor averaged over all other factors is known as a
    main effect, and the effect of a combination of some of the factors
    averaged over all other factors is known as an interaction.
    This can be represented by a linear model.
    In the above example if the response was y_{ijk} for the kth
    replicate of the ith level of T and the jth level of P the linear
    model would be

        y_{ijk} = mu+t_i+p_j+gamma_{ij}+e_{ijk}

    where mu is the overall mean, t_i is the main effect of T, p_j is
    the main effect of P, gamma_{ij} is the T*P interaction and e_{ijk}
    is the random error term.
    In order to find unique estimates constraints are placed on the
    parameters estimates.
    For the example here these are:

        [table omitted]

    where ^ denotes the estimate.

    If there is variation in the experimental conditions (e.g., in an
    experiment on the production of a material different batches of raw
    material may be used, or the experiment may be carried out on
    different days), then plots that are similar are grouped together
    into blocks.
    For a balanced complete factorial experiment all the treatment
    combinations occur the same number of times in each block.

    ``factorial`` computes the analysis of variance (ANOVA) table by
    sequentially computing the totals and means for an effect from the
    residuals computed when previous effects have been removed.
    The effect sum of squares is the sum of squared totals divided by
    the number of observations per total.
    The means are then subtracted from the residuals to compute a new
    set of residuals.
    At the same time the means for the original data are computed.
    When all effects are removed the residual sum of squares is computed
    from the residuals.
    Given the sums of squares an ANOVA table is then computed along with
    standard errors for the difference in treatment means.

    The data for ``factorial`` has to be in standard order given by the
    order of the factors.
    Let there be k factors, f_1,f_2,...,f_k in that order with levels
    l_1,l_2,...,l_k respectively.
    Standard order requires the levels of factor f_1 are in order
    1,2,...,l_1 and within each level of f_1 the levels of f_2 are in
    order 1,2,...,l_2 and so on.

    For an experiment with blocks the data is for block 1 then for block
    2, etc.
    Within each block the data must be arranged so that the levels of
    factor f_1 are in order 1,2,...,l_1 and within each level of f_1 the
    levels of f_2 are in order 1,2,...,l_2 and so on.
    Any within block replication of treatment combinations must occur
    within the levels of f_k.

    The ANOVA table is given in the following order.
    For a complete factorial experiment the first row is for blocks, if
    present, then the main effects of the factors in their order, e.g.,
    f_1 followed by f_2, etc.
    These are then followed by all the two factor interactions then all
    the three factor interactions, etc., the last two rows being for the
    residual and total sums of squares.
    The interactions are arranged in lexical order for the given factor
    order.
    For example, for the three factor interactions for a five factor
    experiment the 10 interactions would be in the following order:

        f_1 f_2 f_3;f_1 f_2 f_4;f_1 f_2 f_5;f_1 f_3 f_4;f_1 f_3 f_5;f_1
        f_4 f_5;f_2 f_3 f_4;f_2 f_3 f_5;f_2 f_4 f_5;f_3 f_4 f_5

    References
    ----------
    Cochran, W G and Cox, G M, 1957, `Experimental Designs`, Wiley

    Davis, O L, 1978, `The Design and Analysis of Industrial
    Experiments`, Longman

    John, J A and Quenouille, M H, 1977, `Experiments: Design and
    Analysis`, Griffin
    """

    excs = []
    fun_name = 'naginterfaces.base.anova.factorial'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g04caft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._BYREF_INT, # Input, nfac.
        utils._EngineIntArrayType.c_type, # Input, lfac[nfac].
        utils._BYREF_INT, # Input, nblock.
        utils._BYREF_INT, # Input, inter.
        utils._BYREF_INT, # Input, irdf.
        utils._BYREF_INT, # Input, mterm.
        utils._EngineFloat64ArrayType.c_type, # Output, table[mterm, 5].
        utils._BYREF_INT, # Output, itotal.
        utils._EngineFloat64ArrayType.c_type, # Output, tmean[maxt].
        utils._BYREF_INT, # Input, maxt.
        utils._EngineFloat64ArrayType.c_type, # Output, e[maxt].
        utils._EngineIntArrayType.c_type, # Output, imean[mterm].
        utils._EngineFloat64ArrayType.c_type, # Output, semean[mterm].
        utils._EngineFloat64ArrayType.c_type, # Output, bmean[nblock+1].
        utils._EngineFloat64ArrayType.c_type, # Output, r[n].
        utils._EngineIntArrayType.c_type, # Workspace, iwk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((table, 'table'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _nfac = utils._EngineIntScalarType(
        nfac, {'fun_name': fun_name, 'entity': 'nfac'},
    ).to_ctypes()
    _nfac_py = utils._EngineIntScalarType.to_py(_nfac)
    _lfac = utils._EngineIntArrayType(
        lfac, {'fun_name': fun_name, 'entity': 'lfac'},
        exp_shape=(_nfac_py,),
    ).to_ctypes()
    _nblock = utils._EngineIntScalarType(
        nblock, {'fun_name': fun_name, 'entity': 'nblock'},
    ).to_ctypes()
    _inter = utils._EngineIntScalarType(
        inter, {'fun_name': fun_name, 'entity': 'inter'},
    ).to_ctypes()
    _irdf = utils._EngineIntScalarType(
        irdf, {'fun_name': fun_name, 'entity': 'irdf'},
    ).to_ctypes()
    _mterm = utils._EngineIntScalarType(
        mterm, {'fun_name': fun_name, 'entity': 'mterm'},
    ).to_ctypes()
    _mterm_py = utils._EngineIntScalarType.to_py(_mterm)
    _table_cls = utils._EngineFloat64ArrayType(
        table, {'fun_name': fun_name, 'entity': 'table'},
        exp_shape=(_mterm_py, 5),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _table = _table_cls.to_ctypes(can_reference_contents=False)
    _itotal = utils._EngineIntScalarType.empty_ctypes()
    _maxt = utils._EngineIntScalarType(
        maxt, {'fun_name': fun_name, 'entity': 'maxt'},
    ).to_ctypes()
    _maxt_py = utils._EngineIntScalarType.to_py(_maxt)
    _tmean_cls = utils._EngineFloat64ArrayType(
        tmean, {'fun_name': fun_name, 'entity': 'tmean'},
        exp_shape=(_maxt_py,),
        check_contents=False,
    )
    _tmean = _tmean_cls.to_ctypes(can_reference_contents=False)
    _e_cls = utils._EngineFloat64ArrayType(
        e, {'fun_name': fun_name, 'entity': 'e'},
        exp_shape=(_maxt_py,),
        check_contents=False,
    )
    _e = _e_cls.to_ctypes(can_reference_contents=False)
    _imean_cls = utils._EngineIntArrayType(
        imean, {'fun_name': fun_name, 'entity': 'imean'},
        exp_shape=(_mterm_py,),
        check_contents=False,
    )
    _imean = _imean_cls.to_ctypes(can_reference_contents=False)
    _semean_cls = utils._EngineFloat64ArrayType(
        semean, {'fun_name': fun_name, 'entity': 'semean'},
        exp_shape=(_mterm_py,),
        check_contents=False,
    )
    _semean = _semean_cls.to_ctypes(can_reference_contents=False)
    _nblock_py = utils._EngineIntScalarType.to_py(_nblock)
    _bmean_cls = utils._EngineFloat64ArrayType(
        bmean, {'fun_name': fun_name, 'entity': 'bmean'},
        exp_shape=(_nblock_py+1,),
        check_contents=False,
    )
    _bmean = _bmean_cls.to_ctypes(can_reference_contents=False)
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _r = _r_cls.to_ctypes(can_reference_contents=False)
    _iwk = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _y,
        _nfac,
        _lfac,
        _nblock,
        _inter,
        _irdf,
        _mterm,
        _table,
        _itotal,
        _tmean,
        _maxt,
        _e,
        _imean,
        _semean,
        _bmean,
        _r,
        _iwk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irdf = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irdf >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, inter = {:s} and nfac = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: inter <= nfac.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, inter = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: inter >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nblock = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nblock >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nfac = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nfac >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and lfac[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lfac[i-1] >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the number of plots per block is incompatible',
                    'with the number of plot factors.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and nblock = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: when nblock > 1, n must be a multiple of nblock.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineNoErrorExit()
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineNoErrorExit()
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the values of y are constant.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99988:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The residual sum of squares is zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99989:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'There are no degrees of freedom for the residual sum of squares.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _table_cls.refresh_py(_table)
        _tmean_cls.refresh_py(_tmean)
        _e_cls.refresh_py(_e)
        _imean_cls.refresh_py(_imean)
        _semean_cls.refresh_py(_semean)
        _bmean_cls.refresh_py(_bmean)
        _r_cls.refresh_py(_r)
        _itotal_py = utils._EngineIntScalarType.to_py(_itotal)
        return_data = _itotal_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['itotal'],
            )(
                _itotal_py,
            )
    _err_class.manage()
    return return_data

def contrasts(nt, tmean, irep, rms, rdf, nc, ct, est, tabl, tol, usetx, tx):
    u"""
    Computes sum of squares for contrast between means.

    ``contrasts`` computes sum of squares for a user-defined contrast
    between means.

    For full information please refer to the NAG Library document for
    g04da

    https://www.nag.com/numeric/nl/nagdoc_27.3/flhtml/g04/g04daf.html

    Parameters
    ----------
    nt : int
        t, the number of treatment means.

    tmean : float, shape (`nt`)
        The treatment means, tau^_i, for i = 1,2,...,t.

    irep : int, shape (`nt`)
        The replication for each treatment mean, n_i, for i = 1,2,...,t.

    rms : float
        The residual mean square, sigma^^2.

    rdf : float
        The residual degrees of freedom.

    nc : int
        The number of contrasts.

    ct : float, shape (`nt`, `nc`)
        The columns of `ct` must contain the `nc` contrasts, that is
        `ct`[i-1,j-1] must contain lambda_{ij}, for j = 1,2,...,`nc`,
        for i = 1,2,...,t.

    est : float, shape (`nc`), modified in place
        `On exit`: the estimates of the contrast, Lambda_j, for j =
        1,2,...,`nc`.

    tabl : float, shape (`nc`, 5), modified in place
        `On entry`: the elements of `tabl` that are not referenced as
        described below remain unchanged.

        `On exit`: the rows of the analysis of variance table for the
        contrasts. For each row column 1 contains the degrees of
        freedom, column 2 contains the sum of squares, column 3 contains
        the mean square, column 4 the F-statistic and column 5 the
        significance level for the contrast. Note that the degrees of
        freedom are always one and so the mean square equals the sum of
        squares.

    tol : float
        The tolerance, epsilon used to check if the contrasts are
        orthogonal and if they are orthogonal to the mean. If `tol` <=
        0.0 the value machine precision is used.

    usetx : bool
        If `usetx` = True the means tau_i^* are provided in `tx` and the
        formula (2) is used instead of formula (1).

        If `usetx` = False formula (1) is used and `tx` is not
        referenced.

    tx : float, shape (`nt`)
        If `usetx` = True `tx` must contain the means tau_i^*, for i =
        1,2,...,t.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `rdf` = *<value>*.

            Constraint: `rdf` >= 1.0.

        (`errno` 1)
            On entry, `rms` = *<value>*.

            Constraint: `rms` > 0.0.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` >= 2.

        (`errno` 1)
            On entry, `nc` = *<value>*.

            Constraint: `nc` >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The *<value>**<value>* and *<value>**<value>* contrasts are
            not orthogonal.

        (`errno` 2)
            The *<value>**<value>* contrast is not orthogonal to the
            mean.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    In the analysis of designed experiments the first stage is to
    compute the basic analysis of variance table, the estimate of the
    error variance (the residual or error mean square), sigma^^2, and
    the (variance ratio) F-statistic for the t treatments.
    If this F-test is significant then the second stage of the analysis
    is to explore which treatments are significantly different.

    If there is a structure to the treatments then this may lead to
    hypotheses that can be defined before the analysis and tested using
    linear contrasts.
    For example, if the treatments were three different fixed
    temperatures, say 18, 20 and 22, and an uncontrolled temperature
    (denoted by N) then the following contrasts might be of interest.

        [table omitted]

    The first represents the average difference between the controlled
    temperatures and the uncontrolled temperature.
    The second represents the linear effect of an increasing fixed
    temperature.

    For a randomized complete block design or a completely randomized
    design, let the treatment means be tau^_i, i = 1,2,...,t, and let
    the jth contrast be defined by lambda_{ij}, i = 1,2,...,t, then the
    estimate of the contrast is simply:

        Lambda_j = sum_i = 1^t tau^_ilambda_{ij}

    and the sum of squares for the contrast is:

        SS_j = Lambda_j^2/(sum_i = 1^t lambda_ij^2/n_i)

    where n_i is the number of observations for the ith treatment.
    Such a contrast has one degree of freedom so that the appropriate
    F-statistic is SS_j/sigma^^2.

    The two contrasts lambda_{ij} and lambda_{ij'} are orthogonal if
    sum_i = 1^t lambda_{ij}lambda_{ij'} = 0 and the contrast lambda_{ij}
    is orthogonal to the overall mean if sum_i = 1^t lambda_{ij} = 0.
    In practice these sums will be tested against a small quantity,
    epsilon.
    If each of a set of contrasts is orthogonal to the mean and they are
    all mutually orthogonal then the contrasts provide a partition of
    the treatment sum of squares into independent components.
    Hence the resulting F-tests are independent.

    If the treatments come from a design in which treatments are not
    orthogonal to blocks then the sum of squares for a contrast is given
    by:

        SS_j = Lambda_jLambda_j^*/(sum_i = 1^t lambda_ij^2/n_i)

    where

        Lambda_j^* = sum_i = 1^t tau_i^*lambda_{ij}

    with tau_i^*, for i = 1,2,...,t, being adjusted treatment means
    computed by first eliminating blocks then computing the treatment
    means from the block adjusted observations without taking into
    account the non-orthogonality between treatments and blocks.
    For further details see John (1987).

    References
    ----------
    Cochran, W G and Cox, G M, 1957, `Experimental Designs`, Wiley

    John, J A, 1987, `Cyclic Designs`, Chapman and Hall

    Winer, B J, 1970, `Statistical Principles in Experimental Design`,
    McGraw--Hill
    """

    excs = []
    fun_name = 'naginterfaces.base.anova.contrasts'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g04daft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nt.
        utils._EngineFloat64ArrayType.c_type, # Input, tmean[nt].
        utils._EngineIntArrayType.c_type, # Input, irep[nt].
        utils._BYREF_FLOAT64, # Input, rms.
        utils._BYREF_FLOAT64, # Input, rdf.
        utils._BYREF_INT, # Input, nc.
        utils._EngineFloat64ArrayType.c_type, # Input, ct[ldct, nc].
        utils._BYREF_INT, # Input, ldct.
        utils._EngineFloat64ArrayType.c_type, # Output, est[nc].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, tabl[ldtabl, 5].
        utils._BYREF_INT, # Input, ldtabl.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_LOGICAL, # Input, usetx.
        utils._EngineFloat64ArrayType.c_type, # Input, tx[nt].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ct, 'ct'), (tabl, 'tabl')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _nt = utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _nt_py = utils._EngineIntScalarType.to_py(_nt)
    _tmean = utils._EngineFloat64ArrayType(
        tmean, {'fun_name': fun_name, 'entity': 'tmean'},
        exp_shape=(_nt_py,),
    ).to_ctypes()
    _irep = utils._EngineIntArrayType(
        irep, {'fun_name': fun_name, 'entity': 'irep'},
        exp_shape=(_nt_py,),
    ).to_ctypes()
    _rms = utils._EngineFloat64ScalarType(
        rms, {'fun_name': fun_name, 'entity': 'rms'},
    ).to_ctypes()
    _rdf = utils._EngineFloat64ScalarType(
        rdf, {'fun_name': fun_name, 'entity': 'rdf'},
    ).to_ctypes()
    _nc = utils._EngineIntScalarType(
        nc, {'fun_name': fun_name, 'entity': 'nc'},
    ).to_ctypes()
    _ldct_py = utils._EngineFloat64ArrayType.get_nd(ct, {'fun_name': fun_name, 'entity': 'ct'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldct = utils._EngineIntScalarType(
        _ldct_py, {'fun_name': fun_name, 'entity': '_ldct'},
    ).to_ctypes()
    _nc_py = utils._EngineIntScalarType.to_py(_nc)
    _ct = utils._EngineFloat64ArrayType(
        ct, {'fun_name': fun_name, 'entity': 'ct'},
        exp_shape=(_nt_py, _nc_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _est_cls = utils._EngineFloat64ArrayType(
        est, {'fun_name': fun_name, 'entity': 'est'},
        exp_shape=(_nc_py,),
        check_contents=False,
    )
    _est = _est_cls.to_ctypes(can_reference_contents=False)
    _ldtabl_py = utils._EngineFloat64ArrayType.get_nd(tabl, {'fun_name': fun_name, 'entity': 'tabl'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldtabl = utils._EngineIntScalarType(
        _ldtabl_py, {'fun_name': fun_name, 'entity': '_ldtabl'},
    ).to_ctypes()
    _tabl_cls = utils._EngineFloat64ArrayType(
        tabl, {'fun_name': fun_name, 'entity': 'tabl'},
        exp_shape=(_nc_py, 5),
        mutual_sorder=_m_sorder,
    )
    _tabl = _tabl_cls.to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _usetx = utils._EngineLogicalScalarType(
        usetx, {'fun_name': fun_name, 'entity': 'usetx'},
    ).to_ctypes()
    _tx = utils._EngineFloat64ArrayType(
        tx, {'fun_name': fun_name, 'entity': 'tx'},
        exp_shape=(_nt_py,),
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nt,
        _tmean,
        _irep,
        _rms,
        _rdf,
        _nc,
        _ct,
        _ldct,
        _est,
        _tabl,
        _ldtabl,
        _tol,
        _usetx,
        _tx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rdf = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rdf >= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rms = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rms > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nt >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nc = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nc >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The {:s}{:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    '{:s}{:s} contrasts are not orthogonal.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The {:s}{:s} contrast is not'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    'orthogonal to the mean.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _est_cls.refresh_py(_est)
        _tabl_cls.refresh_py(_tabl)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def confidence(typ, nt, tmean, rdf, c, clevel, cil, ciu, isig):
    u"""
    Computes confidence intervals for differences between means computed
    by ``random`` or ``rowcol``.

    ``confidence`` computes simultaneous confidence intervals for the
    differences between means.
    It is intended for use after ``random`` or ``rowcol``.

    For full information please refer to the NAG Library document for
    g04db

    https://www.nag.com/numeric/nl/nagdoc_27.3/flhtml/g04/g04dbf.html

    Parameters
    ----------
    typ : str, length 1
        Indicates which method is to be used.

        `typ` = 'T'

            The Tukey--Kramer method is used.

        `typ` = 'B'

            The Bonferroni method is used.

        `typ` = 'D'

            The Dunn--Sidak method is used.

        `typ` = 'L'

            The Fisher LSD method is used.

        `typ` = 'S'

            The Scheffe's method is used.

    nt : int
        t, the number of treatment means.

    tmean : float, shape (`nt`)
        The treatment means, tau^_i, for i = 1,2,...,t.

    rdf : float
        nu, the residual degrees of freedom.

    c : float, shape (`nt`, `nt`)
        The strictly lower triangular part of `c` must contain the
        standard errors of the differences between the means as returned
        by ``random`` and ``rowcol``. That is `c`[i,j], i > j, contains
        the standard error of the difference between the ith and jth
        mean in `tmean`.

    clevel : float
        The required confidence level for the computed intervals,
        (1-alpha).

    cil : float, shape (`nt`*(`nt`-1)/2), modified in place
        `On exit`: the ((i-1)(i-2)/2+j)th element contains the lower
        limit to the confidence interval for the difference between ith
        and jth means in `tmean`, for j = 1,2,...,i-1, for i =
        2,3,...,t.

    ciu : float, shape (`nt`*(`nt`-1)/2), modified in place
        `On exit`: the ((i-1)(i-2)/2+j)th element contains the upper
        limit to the confidence interval for the difference between ith
        and jth means in `tmean`, for j = 1,2,...,i-1, for i =
        2,3,...,t.

    isig : int, shape (`nt`*(`nt`-1)/2), modified in place
        `On exit`: the ((i-1)(i-2)/2+j)th element indicates if the
        difference between ith and jth means in `tmean` is significant,
        for j = 1,2,...,i-1, for i = 2,3,...,t. If the difference is
        significant then the returned value is 1; otherwise the returned
        value is 0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `typ` = *<value>*.

            Constraint: `typ` = 'T', 'B', 'D', 'L' or 'S'.

        (`errno` 1)
            On entry, `clevel` = *<value>*.

            Constraint: 0.0 < `clevel` < 1.0.

        (`errno` 1)
            On entry, `rdf` = *<value>*.

            Constraint: `rdf` >= 1.0.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` >= 2.

        (`errno` 2)
            On entry, i = *<value>* and j = *<value>*.

            Constraint: `c`[i-1,j-1] > 0.0.

        (`errno` 3)
            There has been a failure in the computation of the
            studentized range statistic.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    In the computation of analysis of a designed experiment the first
    stage is to compute the basic analysis of variance table, the
    estimate of the error variance (the residual or error mean square),
    sigma^^2, the residual degrees of freedom, nu, and the (variance
    ratio) F-statistic for the t treatments.
    The second stage of the analysis is to compare the treatment means.
    If the treatments have no structure, for example the treatments are
    different varieties, rather than being structured, for example a set
    of different temperatures, then a multiple comparison procedure can
    be used.

    A multiple comparison procedure looks at all possible pairs of means
    and either computes confidence intervals for the difference in means
    or performs a suitable test on the difference.
    If there are t treatments then there are t(t-1)/2 comparisons to be
    considered.
    In tests the type 1 error or significance level is the probability
    that the result is considered to be significant when there is no
    difference in the means.
    If the usual t-test is used with, say, a 6% significance level then
    the type 1 error for all k = t(t-1)/2 tests will be much higher.
    If the tests were independent then if each test is carried out at
    the 100 alpha percent level then the overall type 1 error would be
    alpha* = 1-(1-alpha)^k ~= k alpha.
    In order to provide an overall protection the individual tests, or
    confidence intervals, would have to be carried out at a value of
    alpha such that alpha* is the required significance level, e.g.,
    five percent.

    The 100*(1-alpha) percent confidence interval for the difference in
    two treatment means, tau^_i and tau^_j is given by

        (tau^_i-tau^_j)+/-T_(alpha,nu,t)^*se(tau^_i-tau^_j),

    where se() denotes the standard error of the difference in means and
    T_(alpha,nu,t)^* is an appropriate percentage point from a
    distribution.
    There are several possible choices for T_(alpha,nu,t)^*.
    These are:

    (a) 1/2q_(1-alpha,nu,t), the studentized range statistic, see
        ``stat.inv_cdf_studentized_range``. It is the appropriate
        statistic to compare the largest mean with the smallest mean.
        This is known as Tukey--Kramer method.

    (#) t_(alpha/k,nu), this is the Bonferroni method.

    (#) t_(alpha_0,nu), where alpha_0 = 1-(1-alpha)^(1/k), this is known
        as the Dunn--Sidak method.

    (#) t_(alpha,nu), this is known as Fisher's LSD (least significant
        difference) method. It should only be used if the overall F-test
        is significant, the number of treatment comparisons is small and
        were planned before the analysis.

    (#) sqrt((k-1)F_{1-alpha,k-1,nu}) where F_{1-alpha,k-1,nu} is the
        deviate corresponding to a lower tail probability of 1-alpha
        from an F-distribution with k-1 and nu degrees of freedom. This
        is Scheffe's method.

    In cases (b), (c) and (d), t_(alpha,nu) denotes the alpha two tail
    significance level for the Student's t-distribution with nu degrees
    of freedom, see ``stat.inv_cdf_students_t``.

    The Scheffe method is the most conservative, followed closely by the
    Dunn--Sidak and Tukey--Kramer methods.

    To compute a test for the difference between two means the
    statistic,

        (tau^_i-tau^_j)/(se(tau^_i-tau^_j))

    is compared with the appropriate value of T_(alpha,nu,t)^*.

    References
    ----------
    Kotz, S and Johnson, N L (ed.), 1985, `Multiple range and associated
    test procedures`, Encyclopedia of Statistical Sciences (5), Wiley,
    New York

    Kotz, S and Johnson, N L (ed.), 1985, `Multiple comparison`,
    Encyclopedia of Statistical Sciences (5), Wiley, New York

    Winer, B J, 1970, `Statistical Principles in Experimental Design`,
    McGraw--Hill
    """

    excs = []
    fun_name = 'naginterfaces.base.anova.confidence'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g04dbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, typ(len=1).
        utils._BYREF_INT, # Input, nt.
        utils._EngineFloat64ArrayType.c_type, # Input, tmean[nt].
        utils._BYREF_FLOAT64, # Input, rdf.
        utils._EngineFloat64ArrayType.c_type, # Input, c[ldc, nt].
        utils._BYREF_INT, # Input, ldc.
        utils._BYREF_FLOAT64, # Input, clevel.
        utils._EngineFloat64ArrayType.c_type, # Output, cil[nt*(nt-1)//2].
        utils._EngineFloat64ArrayType.c_type, # Output, ciu[nt*(nt-1)//2].
        utils._EngineIntArrayType.c_type, # Output, isig[nt*(nt-1)//2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of typ.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((c, 'c'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _typ = utils._EngineCharScalarType(
        typ, {'fun_name': fun_name, 'entity': 'typ'}, exp_chrlen=1,
    ).to_ctypes()
    _nt = utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _nt_py = utils._EngineIntScalarType.to_py(_nt)
    _tmean = utils._EngineFloat64ArrayType(
        tmean, {'fun_name': fun_name, 'entity': 'tmean'},
        exp_shape=(_nt_py,),
    ).to_ctypes()
    _rdf = utils._EngineFloat64ScalarType(
        rdf, {'fun_name': fun_name, 'entity': 'rdf'},
    ).to_ctypes()
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_nt_py, _nt_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _clevel = utils._EngineFloat64ScalarType(
        clevel, {'fun_name': fun_name, 'entity': 'clevel'},
    ).to_ctypes()
    _cil_cls = utils._EngineFloat64ArrayType(
        cil, {'fun_name': fun_name, 'entity': 'cil'},
        exp_shape=(_nt_py*(_nt_py-1)//2,),
        check_contents=False,
    )
    _cil = _cil_cls.to_ctypes(can_reference_contents=False)
    _ciu_cls = utils._EngineFloat64ArrayType(
        ciu, {'fun_name': fun_name, 'entity': 'ciu'},
        exp_shape=(_nt_py*(_nt_py-1)//2,),
        check_contents=False,
    )
    _ciu = _ciu_cls.to_ctypes(can_reference_contents=False)
    _isig_cls = utils._EngineIntArrayType(
        isig, {'fun_name': fun_name, 'entity': 'isig'},
        exp_shape=(_nt_py*(_nt_py-1)//2,),
        check_contents=False,
    )
    _isig = _isig_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _typ,
        _nt,
        _tmean,
        _rdf,
        _c,
        _ldc,
        _clevel,
        _cil,
        _ciu,
        _isig,
        _errbuf,
        _ierr,
        _typ._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, typ = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: typ = \'T\', \'B\', \'D\', \'L\' or \'S\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, clevel = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < clevel < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rdf = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rdf >= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nt >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and j = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: c[i-1,j-1] > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'There has been a failure in the computation of',
                    'the studentized range statistic.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _cil_cls.refresh_py(_cil)
        _ciu_cls.refresh_py(_ciu)
        _isig_cls.refresh_py(_isig)
        return_data = None
    _err_class.manage()
    return return_data

def dummyvars(typ, n, levels, ifact, x, v, rep):
    u"""
    Computes orthogonal polynomials or dummy variables for
    factor/classification variable.

    ``dummyvars`` computes orthogonal polynomial or dummy variables for
    a factor or classification variable.

    For full information please refer to the NAG Library document for
    g04ea

    https://www.nag.com/numeric/nl/nagdoc_27.3/flhtml/g04/g04eaf.html

    Parameters
    ----------
    typ : str, length 1
        The type of dummy variable to be computed.

            If `typ` = 'P', an orthogonal Polynomial representation is
            computed.

            If `typ` = 'H', a Helmert matrix representation is computed.

            If `typ` = 'F', the contrasts relative to the First level
            are computed.

            If `typ` = 'L', the contrasts relative to the Last level are
            computed.

            If `typ` = 'C', a Complete set of dummy variables is
            computed.

    n : int
        n, the number of observations for which the dummy variables are
        to be computed.

    levels : int
        k, the number of levels of the factor.

    ifact : int, shape (`n`)
        The n values of the factor.

    x : float, shape (`n`, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `typ` in ('P', 'H', 'F', 'L'):
        `levels`-1; if `typ`='C': `levels`; otherwise: 0.

        `On exit`: the n*k* matrix of dummy variables, where k* = k-1 if
        `typ` = 'P', 'H', 'F' or 'L' and k* = k if `typ` = 'C'.

    v : float, shape (:)
        Note: the required length for this argument is determined as
        follows: if `typ`='P': `levels`; otherwise: 1.

        If `typ` = 'P', the k distinct values of the underlying variable
        for which the orthogonal polynomial is to be computed.

        If `typ` != 'P', `v` is not referenced.

    rep : float, shape (`levels`), modified in place
        `On exit`: the number of replications for each level of the
        factor, r_i, for i = 1,2,...,k.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `typ` = *<value>*.

            Constraint: `typ` = 'P', 'H', 'F', 'L' or 'C'.

        (`errno` 1)
            On entry, `n` = *<value>* and `levels` = *<value>*.

            Constraint: `n` >= `levels`.

        (`errno` 1)
            On entry, `levels` = *<value>*.

            Constraint: `levels` >= 2.

        (`errno` 2)
            On entry, not all values of `v` are distinct.

        (`errno` 2)
            On entry, not all levels are present in `ifact`.

        (`errno` 2)
            On entry, i = *<value>*, `ifact`[i-1] = *<value>* and
            `levels` = *<value>*.

            Constraint: 1 <= `levels`[i-1] <= `levels`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The *<value>**<value>* polynomial has all elements zero.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    In the analysis of an experimental design using a general linear
    model the factors or classification variables that specify the
    design have to be coded as dummy variables. ``dummyvars`` computes
    dummy variables that can then be used in the fitting of the general
    linear model using ``correg.linregm_fit``.

    If the factor of length n has k levels then the simplest
    representation is to define k dummy variables, X_j such that X_j = 1
    if the factor is at level j and 0 otherwise for j = 1,2,...,k.
    However, there is usually a mean included in the model and the sum
    of the dummy variables will be aliased with the mean.
    To avoid the extra redundant argument k-1 dummy variables can be
    defined as the contrasts between one level of the factor, the
    reference level, and the remaining levels.
    If the reference level is the first level then the dummy variables
    can be defined as X_j = 1 if the factor is at level j and 0
    otherwise, for j = 2,3,...,k.
    Alternatively, the last level can be used as the reference level.

    A second way of defining the k-1 dummy variables is to use a Helmert
    matrix in which levels 2,3,...,k are compared with the average
    effect of the previous levels.
    For example if k = 4 then the contrasts would be:

        [table omitted]

    Thus variable j, for j = 1,2,...,k-1 is given by

        X_j = -1 if factor is at level less than j+1

        X_j = sum_i = 1^jr_i/r_{j+1} if factor is at level j+1

        X_j = 0 if factor is at level greater than j+1

    where r_j is the number of replicates of level j.

    If the factor can be considered as a set of values from an
    underlying continuous variable then the factor can be represented by
    a set of k-1 orthogonal polynomials representing the linear,
    quadratic etc. effects of the underlying variable.
    The orthogonal polynomial is computed using Forsythe's algorithm
    (Forsythe (1957), see also Cooper (1968)).
    The values of the underlying continuous variable represented by the
    factor levels have to be supplied to the function.

    The orthogonal polynomials are standardized so that the sum of
    squares for each dummy variable is one.
    For the other methods integer (+/-1) representations are retained
    except that in the Helmert representation the code of level j+1 in
    dummy variable j will be a fraction.

    References
    ----------
    Cooper, B E, 1968, `Algorithm AS 10. The use of orthogonal
    polynomials`, Appl. Statist. (17), 283--287

    Forsythe, G E, 1957, `Generation and use of orthogonal polynomials
    for data fitting with a digital computer`, J. Soc. Indust. Appl.
    Math. (5), 74--88
    """

    excs = []
    fun_name = 'naginterfaces.base.anova.dummyvars'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g04eaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, typ(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, levels.
        utils._EngineIntArrayType.c_type, # Input, ifact[n].
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, *].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineFloat64ArrayType.c_type, # Input, v[*].
        utils._EngineFloat64ArrayType.c_type, # Output, rep[levels].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of typ.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _typ = utils._EngineCharScalarType(
        typ, {'fun_name': fun_name, 'entity': 'typ'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _levels = utils._EngineIntScalarType(
        levels, {'fun_name': fun_name, 'entity': 'levels'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ifact = utils._EngineIntArrayType(
        ifact, {'fun_name': fun_name, 'entity': 'ifact'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _typ_py = utils._EngineCharScalarType.to_py(_typ)
    _levels_py = utils._EngineIntScalarType.to_py(_levels)
    if _typ_py in ('P', 'p', 'H', 'h', 'F', 'f', 'L', 'l'):
        _x_dim2 = _levels_py-1
    elif _typ_py in ('C', 'c'):
        _x_dim2 = _levels_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_n_py, _x_dim2)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    if _typ_py in ('P', 'p'):
        _v_dim1 = _levels_py
    else:
        _v_dim1 = 1
    exp_shape_v = (_v_dim1,)
    _v = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=exp_shape_v,
    ).to_ctypes()
    _rep_cls = utils._EngineFloat64ArrayType(
        rep, {'fun_name': fun_name, 'entity': 'rep'},
        exp_shape=(_levels_py,),
        check_contents=False,
    )
    _rep = _rep_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _typ,
        _n,
        _levels,
        _ifact,
        _x,
        _ldx,
        _v,
        _rep,
        _errbuf,
        _ierr,
        _typ._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, typ = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: typ = \'P\', \'H\', \'F\', \'L\' or \'C\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and levels = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n >= levels.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, levels = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: levels >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, not all values of v are distinct.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, not all levels are present in ifact.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, ifact[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'levels = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= levels[i-1] <= levels.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The {:s}{:s} polynomial has all elements'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    'zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        _rep_cls.refresh_py(_rep)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def icc(mtype, rtype, nrep, nsubj, nrater, score, smiss, alpha):
    u"""
    Intraclass correlation (ICC) for assessing rater reliability.

    ``icc`` calculates the intraclass correlation (ICC).

    For full information please refer to the NAG Library document for
    g04ga

    https://www.nag.com/numeric/nl/nagdoc_27.3/flhtml/g04/g04gaf.html

    Parameters
    ----------
    mtype : int
        Indicates which model is to be used.

        `mtype` = 1

            The reliability study is a one-factor design, ICC(1,1).

        `mtype` = 2

            The reliability study is a random factorial design,
            ICC(2,1).

        `mtype` = 3

            The reliability study is a mixed factorial design, ICC(3,1).

    rtype : int
        Indicates which type of reliability is required.

        `rtype` = 1

            Interrater reliability is required.

        `rtype` = 2

            Intrarater reliability is required.

    nrep : int
        m, the number of replicates.

    nsubj : int
        n_s, the number of subjects.

    nrater : int
        n_r, the number of raters.

    score : float, shape (`nrep`, `nsubj`, `nrater`)
        The matrix of scores, with `score`[k-1,i-1,j-1] being the score
        given to the ith subject by the jth rater in the kth replicate.

        If rater j did not rate subject i at replication k, the
        corresponding element of `score`, `score`[k-1,i-1,j-1], should
        be set to `smiss`.

    smiss : None or float
        The value used to indicate a missing score.

        Care should be taken in the selection of the value used to
        indicate a missing score. ``icc`` will treat any score in the
        inclusive range (1+/-0.1^(machine.decimal_digits-2))*`smiss` as
        missing.

        Alternatively, a NaN (Not A Number) can be used to indicate
        missing values, in which case the value of `smiss` and any
        missing values of `score` can be set through a call to
        ``ieee.create_nan``.

    alpha : float
        alpha, the significance level used in the construction of the
        confidence intervals for `icc`.

    Returns
    -------
    icc : float
        An estimate of the intraclass correlation to measure either the
        interrater reliability, rho, or intrarater reliability, gamma,
        as specified by `mtype` and `rtype`.

    lci : float
        An approximate lower limit for the 100*(1-alpha)% confidence
        interval for the ICC.

    uci : float
        An approximate upper limit for the 100*(1-alpha%) confidence
        interval for the ICC.

    fstat : float
        f, the F-statistic associated with `icc`.

    df1 : float
        nu_1 and nu_2, the degrees of freedom associated with f.

    df2 : float
        nu_1 and nu_2, the degrees of freedom associated with f.

    pvalue : float
        P(F >= f:nu_1,nu_1), the upper tail probability from an F
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `mtype` = *<value>*.

            Constraint: `mtype` = 1, 2 or 3.

        (`errno` 21)
            On entry, `rtype` = *<value>*.

            Constraint: `rtype` = 1 or 2.

        (`errno` 31)
            On entry, `nrep` = *<value>*.

            Constraint: `nrep` >= 1.

        (`errno` 32)
            On entry, `nrep` = *<value>*.

            Constraint: when `mtype` = 2 or 3 and `rtype` = 2, `nrep` >=
            2.

        (`errno` 33)
            On entry, after adjusting for missing data, `nrep` =
            *<value>*.

            Constraint: `nrep` >= 1.

        (`errno` 34)
            On entry, after adjusting for missing data, `nrep` =
            *<value>*.

            Constraint: when `mtype` = 2 or 3 and `rtype` = 2, `nrep` >=
            2.

        (`errno` 41)
            On entry, `nsubj` = *<value>*.

            Constraint: `nsubj` >= 2.

        (`errno` 42)
            On entry, after adjusting for missing data, `nsubj` =
            *<value>*.

            Constraint: `nsubj` >= 2.

        (`errno` 51)
            On entry, `nrater` = *<value>*.

            Constraint: `nrater` >= 2.

        (`errno` 52)
            On entry, after adjusting for missing data, `nrater` =
            *<value>*.

            Constraint: `nrater` >= 2.

        (`errno` 61)
            Unable to calculate the ICC due to a division by zero.

            This is often due to degenerate data, for example all scores
            being the same.

        (`errno` 71)
            On entry, mscore = *<value>*.

            Constraint: mscore = 1 or 2.

        (`errno` 91)
            On entry, `alpha` = *<value>*.

            Constraint: 0.0 < `alpha` < 1.0.

        (`errno` 92)
            On entry, `alpha` = *<value>*.

            `alpha` is too close to either zero or one.

            This error is unlikely to occur.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 62)
            On entry, a replicate, subject or rater contained all
            missing data.

            All output quantities have been calculated using the reduced
            problem size.

        (`errno` 101)
            `icc` does not fall into the interval [`lci`,`uci`].

            All output quantities have been calculated.

        (`errno` 102)
            The estimate of at least one variance component was
            negative.

            Negative estimates were set to zero and all output
            quantities calculated as documented.

    Notes
    -----
    Many scientific investigations involve assigning a value (score) to
    a number of objects of interest (subjects).
    In most instances the method used to score the subject will be
    affected by measurement error which can affect the analysis and
    interpretation of the data.
    When the score is based on the subjective opinion of one or more
    individuals (raters) the measurement error can be high and,
    therefore, it is important to be able to assess its magnitude.
    One way of doing this is to run a reliability study and calculate
    the intraclass correlation (ICC).

    In a typical reliability study each of a random sample of n_s
    subjects are scored, independently, by n_r raters.
    Each rater scores the same subject m times (i.e., there are m
    replicate scores).
    The scores, y_{ijk}, for k = 1,2,...,m, for j = 1,2,...,n_r, for i =
    1,2,...,n_s can be arranged into m data tables, with the n_s rows of
    the table, labelled 1,2,...,n_s, corresponding to the subjects and
    the n_r columns of the table, labelled 1,2,...,n_r, to the raters.
    For example the following data, taken from Shrout and Fleiss (1979),
    shows a typical situation where four raters (n_r = 4) have scored
    six subjects (n_s = 6) once, i.e., there has been no replication (m
    = 1).

    [table omitted]

    The term intraclass correlation is a general one and can mean either
    a measure of interrater reliability, i.e., a measure of how similar
    the raters are, or intrarater reliability, i.e., a measure of how
    consistent each rater is.

    There are a numerous different versions of the ICC, six of which can
    be calculated using ``icc``.
    The different versions of the ICC can lead to different conclusions
    when applied to the same data, it is, therefore, essential to choose
    the most appropriate based on the design of the reliability study
    and whether inter - or intrarater reliability is of interest.
    The six measures of the ICC are split into three different types of
    studies, denoted: ICC(1,1), ICC(2,1) and ICC(3,1).
    This notation ties up with that used by Shrout and Fleiss (1979).
    Each class of study results in two forms of the ICC, depending on
    whether inter - or intrarater reliability is of interest.

    ICC(1,1): One-Factor Design
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~
    The one-factor designs differ, depending on whether inter - or
    intrarater reliability is of interest:

    Interrater reliability
    ++++++++++++++++++++++
    In a one-factor design to measure interrater reliability, each
    subject is scored by a different set of raters randomly selected
    from a larger population of raters.
    Therefore, even though they use the same set of labels each row of
    the data table is associated with a different set of raters.

    A model of the following form is assumed:

        y_{ijk} = mu+s_i+epsilon_{ijk}

    where s_i is the subject effect and epsilon_{ijk} is the error term,
    with s_i~N(0,sigma_s^2) and epsilon_{ijk}~N(0,sigma_epsilon^2).

    The measure of the interrater reliability, rho, is then given by:

        rho = (sigma^_s^2)/(sigma^_s^2+sigma^_epsilon^2)

    where sigma^_s and sigma^_epsilon are the estimated values of
    sigma_s and sigma_epsilon respectively.

    Intrarater reliability
    ++++++++++++++++++++++
    In a one-factor design to measure intrarater reliability, each rater
    scores a different set of subjects.
    Therefore, even though they use the same set of labels, each column
    of the data table is associated with a different set of subjects.

    A model of the following form is assumed:

        y_{ijk} = mu+r_j+epsilon_{ijk}

    where r_i is the rater effect and epsilon_{ijk} is the error term,
    with r_j~N(0,sigma_r^2) and epsilon_{ijk}~N(0,sigma_epsilon^2).

    The measure of the intrarater reliability, gamma, is then given by:

        gamma = (sigma^_r^2)/(sigma^_r^2+sigma^_epsilon^2)

    where sigma^_r and sigma^_epsilon are the estimated values of
    sigma_r and sigma_epsilon respectively.

    ICC(2,1): Random Factorial Design
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    In a random factorial design, each subject is scored by the same set
    of raters.
    The set of raters have been randomly selected from a larger
    population of raters.

    A model of the following form is assumed:

        y_{ijk} = mu+s_i+r_j+(sr)_{ij}+epsilon_{ijk}

    where s_i is the subject effect, r_i is the rater effect, (sr)_{ij}
    is the subject-rater interaction effect and epsilon_{ijk} is the
    error term, with s_i~N(0,sigma_s^2), r_j~N(0,sigma_r^2),
    (sr)_{ij}~N(0,sigma_sr^2) and epsilon_{ijk}~N(0,sigma_epsilon^2).

    Interrater reliability
    ++++++++++++++++++++++
    The measure of the interrater reliability, rho, is given by:

        rho =
        (sigma^_s^2)/(sigma^_s^2+sigma^_r^2+sigma^_sr^2+sigma^_epsilon^2)

    where sigma^_s, sigma^_r, sigma^_{sr} and sigma^_epsilon are the
    estimated values of sigma_s, sigma_r, sigma_{sr} and sigma_epsilon
    respectively.

    Intrarater reliability
    ++++++++++++++++++++++
    The measure of the intrarater reliability, gamma, is given by:

        gamma =
        (sigma^_r^2)/(sigma^_s^2+sigma^_r^2+sigma^_sr^2+sigma^_epsilon^2)

    where sigma^_s, sigma^_r, sigma^_{sr} and sigma^_epsilon are the
    estimated values of sigma_s, sigma_r, sigma_{sr} and sigma_epsilon
    respectively.

    ICC(3,1): Mixed Factorial Design
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    In a mixed factorial design, each subject is scored by the same set
    of raters and these are the only raters of interest.

    A model of the following form is assumed:

        y_{ijk} = mu+s_i+r_j+(sr)_{ij}+epsilon_{ijk}

    where s_i is the subject effect, r_i is the fixed rater effect,
    (sr)_{ij} is the subject-rater interaction effect and epsilon_{ijk}
    is the error term, with s_i~N(0,sigma_s^2), Σ_j = 1^n_rr_j = 0,
    (sr)_{ij}~N(0,sigma_sr^2), Σ_j = 1^n_r(sr)_{ij} = 0 and
    epsilon_{ijk}~N(0,sigma_epsilon^2).

    Interrater reliability
    ++++++++++++++++++++++
    The measure of the interrater reliability, rho, is then given by:

        rho =
        (sigma^_s^2-sigma^_sr^2/(r-1))/(sigma^_s^2+sigma^_sr^2+sigma^_epsilon^2)

    where sigma^_s, sigma^_{sr} and sigma^_epsilon are the estimated
    values of sigma_s, sigma_{sr} and sigma_epsilon respectively.

    Intrarater reliability
    ++++++++++++++++++++++
    The measure of the intrarater reliability, gamma, is then given by:

        gamma =
        (sigma^_s^2+sigma^_sr^2)/(sigma^_s^2+sigma^_sr^2+sigma^_epsilon^2)

    where sigma^_s, sigma^_{sr} and sigma^_epsilon are the estimated
    values of sigma_s, sigma_{sr} and sigma_epsilon respectively.

    As well as an estimate of the ICC, ``icc`` returns an approximate
    (1-alpha)% confidence interval for the ICC and an F-statistic, f,
    associated degrees of freedom (nu_1 and nu_2) and p-value, p, for
    testing that the ICC is zero.

    Details on the formula used to calculate the confidence interval, f,
    nu_1, nu_2, sigma^_s^2, sigma^_r^2, sigma^_sr^2 and sigma^_epsilon^2
    are given in Gwet (2014).
    In the case where there are no missing data these should tie up with
    the formula presented in Shrout and Fleiss (1979).

    In some circumstances, the formula presented in Gwet (2014) for
    calculating sigma^_s^2, sigma^_r^2, sigma^_sr^2 and sigma^_epsilon^2
    can result in a negative value being calculated.
    In such instances, `errno` = 102, the offending estimate is set to
    zero and the calculations continue as normal.

    It should be noted that Shrout and Fleiss (1979) also present
    methods for calculating the ICC based on average scores, denoted
    ICC(1,k), ICC(2,k) and ICC(3,k).
    These are not supplied here as multiple replications are allowed (m
    > 1) hence there is no need to average the scores prior to
    calculating ICC when using ``icc``.

    References
    ----------
    Gwet, K L, 2014, `Handbook of Inter-rater Reliability`, Fourth
    Edition, Advanced Analytics LLC

    Shrout, P E and Fleiss, J L, 1979, `Intraclass Correlations: Uses in
    Assessing Rater Reliability`, Pyschological Bulletin, Vol 86 (2),
    420--428

    See Also
    --------
    :meth:`naginterfaces.library.examples.anova.icc_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.anova.icc'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g04gaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mtype.
        utils._BYREF_INT, # Input, rtype.
        utils._BYREF_INT, # Input, nrep.
        utils._BYREF_INT, # Input, nsubj.
        utils._BYREF_INT, # Input, nrater.
        utils._EngineFloat64ArrayType.c_type, # Input, score[nrep, nsubj, nrater].
        utils._BYREF_INT, # Input, mscore.
        utils._BYREF_FLOAT64, # Input, smiss.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._BYREF_FLOAT64, # Output, icc.
        utils._BYREF_FLOAT64, # Output, lci.
        utils._BYREF_FLOAT64, # Output, uci.
        utils._BYREF_FLOAT64, # Output, fstat.
        utils._BYREF_FLOAT64, # Output, df1.
        utils._BYREF_FLOAT64, # Output, df2.
        utils._BYREF_FLOAT64, # Output, pvalue.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((score, 'score'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _mtype = utils._EngineIntScalarType(
        mtype, {'fun_name': fun_name, 'entity': 'mtype'},
    ).to_ctypes()
    _rtype = utils._EngineIntScalarType(
        rtype, {'fun_name': fun_name, 'entity': 'rtype'},
    ).to_ctypes()
    _nrep = utils._EngineIntScalarType(
        nrep, {'fun_name': fun_name, 'entity': 'nrep'},
    ).to_ctypes()
    _nsubj = utils._EngineIntScalarType(
        nsubj, {'fun_name': fun_name, 'entity': 'nsubj'},
    ).to_ctypes()
    _nrater = utils._EngineIntScalarType(
        nrater, {'fun_name': fun_name, 'entity': 'nrater'},
    ).to_ctypes()
    _nrep_py = utils._EngineIntScalarType.to_py(_nrep)
    _nsubj_py = utils._EngineIntScalarType.to_py(_nsubj)
    _nrater_py = utils._EngineIntScalarType.to_py(_nrater)
    _score = utils._EngineFloat64ArrayType(
        score, {'fun_name': fun_name, 'entity': 'score'},
        exp_shape=(_nrep_py, _nsubj_py, _nrater_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    if smiss is not None:
        _mscore = utils._EngineIntScalarType.c_type(2)
    else:
        _mscore = utils._EngineIntScalarType.c_type(1)
    if smiss is not None:
        if smiss is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'smiss'},
                'must not be None when smiss is not None',
            )
    if smiss is None:
        _smiss = utils._EngineFloat64ScalarType.c_type(0.0)
    else:
        _smiss = utils._EngineFloat64ScalarType(
            smiss, {'fun_name': fun_name, 'entity': 'smiss'},
            allow_none=True,
        ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _icc = utils._EngineFloat64ScalarType.empty_ctypes()
    _lci = utils._EngineFloat64ScalarType.empty_ctypes()
    _uci = utils._EngineFloat64ScalarType.empty_ctypes()
    _fstat = utils._EngineFloat64ScalarType.empty_ctypes()
    _df1 = utils._EngineFloat64ScalarType.empty_ctypes()
    _df2 = utils._EngineFloat64ScalarType.empty_ctypes()
    _pvalue = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mtype,
        _rtype,
        _nrep,
        _nsubj,
        _nrater,
        _score,
        _mscore,
        _smiss,
        _alpha,
        _icc,
        _lci,
        _uci,
        _fstat,
        _df1,
        _df2,
        _pvalue,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mtype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mtype = 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rtype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rtype = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrep = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrep >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 32 and _ifmt == 32:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrep = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: when mtype = 2 or 3 and rtype = 2, nrep >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 33 and _ifmt == 33:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, after adjusting for missing data, nrep = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrep >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 34 and _ifmt == 34:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, after adjusting for missing data, nrep = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: when mtype = 2 or 3 and rtype = 2, nrep >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nsubj = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nsubj >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 42 and _ifmt == 42:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, after adjusting for missing data, nsubj = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nsubj >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrater = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrater >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 52 and _ifmt == 52:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, after adjusting for missing data, nrater = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrater >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Unable to calculate the ICC due to a division by zero.',
                    'This is often due to degenerate data, for example all scores being the same.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 62 and _ifmt == 62:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a replicate, subject or rater contained all missing data.',
                    'All output quantities have been calculated using the reduced problem size.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 71 and _ifmt == 71:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mscore = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mscore = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 91 and _ifmt == 91:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, alpha = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < alpha < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 92 and _ifmt == 92:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, alpha = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'alpha is too close to either zero or one.',
                    'This error is unlikely to occur.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 101 and _ifmt == 101:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'icc does not fall into the interval [lci,uci].',
                    'All output quantities have been calculated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 102 and _ifmt == 102:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The estimate of at least one variance component was negative.',
                    'Negative estimates were set to zero and all output quantities',
                    'calculated as documented.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _icc_py = utils._EngineFloat64ScalarType.to_py(_icc)
        _lci_py = utils._EngineFloat64ScalarType.to_py(_lci)
        _uci_py = utils._EngineFloat64ScalarType.to_py(_uci)
        _fstat_py = utils._EngineFloat64ScalarType.to_py(_fstat)
        _df1_py = utils._EngineFloat64ScalarType.to_py(_df1)
        _df2_py = utils._EngineFloat64ScalarType.to_py(_df2)
        _pvalue_py = utils._EngineFloat64ScalarType.to_py(_pvalue)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['icc', 'lci', 'uci', 'fstat', 'df1', 'df2', 'pvalue'],
        )(
            _icc_py,
            _lci_py,
            _uci_py,
            _fstat_py,
            _df1_py,
            _df2_py,
            _pvalue_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data
